# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

import os
from .utils import check_if_nparray_and_flatten


INTEGRATOR_TYPES = ('ERK', 'IRK', 'GNSF', 'DISCRETE', 'LIFTED_IRK')
COLLOCATION_TYPES = ('GAUSS_RADAU_IIA', 'GAUSS_LEGENDRE', 'EXPLICIT_RUNGE_KUTTA')
COST_DISCRETIZATION_TYPES = ('EULER', 'INTEGRATOR')

class AcadosOcpOptions:
    """
    class containing the description of the solver options
    """
    def __init__(self):
        self.__hessian_approx = 'GAUSS_NEWTON'
        self.__integrator_type = 'ERK'
        self.__tf = None
        self.__N_horizon = None
        self.__nlp_solver_type = 'SQP'
        self.__nlp_solver_tol_stat = 1e-6
        self.__nlp_solver_tol_eq = 1e-6
        self.__nlp_solver_tol_ineq = 1e-6
        self.__nlp_solver_tol_comp = 1e-6
        self.__nlp_solver_max_iter = 100
        self.__nlp_solver_ext_qp_res = 0
        self.__nlp_solver_warm_start_first_qp = False
        self.__nlp_solver_tol_min_step_norm = None
        self.__globalization = 'FIXED_STEP'
        self.__levenberg_marquardt = 0.0
        self.__collocation_type = 'GAUSS_LEGENDRE'
        self.__sim_method_num_stages = 4
        self.__sim_method_num_steps = 1
        self.__sim_method_newton_iter = 3
        self.__sim_method_newton_tol = 0.0
        self.__sim_method_jac_reuse = 0
        self.__shooting_nodes = None
        self.__time_steps = None
        self.__cost_scaling = None
        self.__Tsim = None
        self.__qp_solver = 'PARTIAL_CONDENSING_HPIPM'
        self.__qp_solver_tol_stat = None
        self.__qp_solver_tol_eq = None
        self.__qp_solver_tol_ineq = None
        self.__qp_solver_tol_comp = None
        self.__qp_solver_iter_max = 50
        self.__qp_solver_cond_N = None
        self.__qp_solver_cond_block_size = None
        self.__qp_solver_warm_start = 0
        self.__qp_solver_cond_ric_alg = 1
        self.__qp_solver_ric_alg = 1
        self.__qp_solver_mu0 = 0.0
        self.__rti_log_residuals = 0
        self.__rti_log_only_available_residuals = 0
        self.__print_level = 0
        self.__cost_discretization = 'EULER'
        self.__regularize_method = 'NO_REGULARIZE'
        self.__reg_epsilon = 1e-4
        self.__exact_hess_cost = 1
        self.__exact_hess_dyn = 1
        self.__exact_hess_constr = 1
        self.__eval_residual_at_max_iter = None
        self.__fixed_hess = 0
        self.__globalization_funnel_init_increase_factor = 15.0
        self.__globalization_funnel_init_upper_bound = 1.0
        self.__globalization_funnel_sufficient_decrease_factor = 0.9
        self.__globalization_funnel_kappa = 0.9
        self.__globalization_funnel_fraction_switching_condition = 1e-3
        self.__globalization_funnel_initial_penalty_parameter = 1.0
        self.__globalization_fixed_step_length = 1.0
        self.__ext_cost_num_hess = 0
        self.__globalization_use_SOC = 0
        self.__globalization_alpha_min = None
        self.__globalization_alpha_reduction = None
        self.__globalization_line_search_use_sufficient_descent = 0
        self.__globalization_full_step_dual = None
        self.__globalization_eps_sufficient_descent = None
        self.__hpipm_mode = 'BALANCE'
        self.__with_solution_sens_wrt_params = False
        self.__with_value_sens_wrt_params = False
        self.__as_rti_iter = 1
        self.__as_rti_level = 4
        self.__with_adaptive_levenberg_marquardt = False
        self.__adaptive_levenberg_marquardt_lam = 5.0
        self.__adaptive_levenberg_marquardt_mu_min = 1e-16
        self.__adaptive_levenberg_marquardt_mu0 = 1e-3
        self.__log_primal_step_norm: bool = False
        self.__store_iterates: bool = False
        self.__timeout_max_time = 0.
        self.__timeout_heuristic = 'LAST'

        # TODO: move those out? they are more about generation than about the acados OCP solver.
        env = os.environ
        self.__ext_fun_compile_flags = '-O2' if 'ACADOS_EXT_FUN_COMPILE_FLAGS' not in env else env['ACADOS_EXT_FUN_COMPILE_FLAGS']
        self.__model_external_shared_lib_dir = None
        self.__model_external_shared_lib_name = None
        self.__custom_update_filename = ''
        self.__custom_update_header_filename = ''
        self.__custom_templates = []
        self.__custom_update_copy = True
        self.__num_threads_in_batch_solve: int = 1

    @property
    def qp_solver(self):
        """QP solver to be used in the NLP solver.
        String in ('PARTIAL_CONDENSING_HPIPM', 'FULL_CONDENSING_QPOASES', 'FULL_CONDENSING_HPIPM', 'PARTIAL_CONDENSING_QPDUNES', 'PARTIAL_CONDENSING_OSQP', 'FULL_CONDENSING_DAQP').
        Default: 'PARTIAL_CONDENSING_HPIPM'.
        """
        return self.__qp_solver

    @property
    def ext_fun_compile_flags(self):
        """
        String with compiler flags for external function compilation.
        Default: '-O2' if environment variable ACADOS_EXT_FUN_COMPILE_FLAGS is not set, else ACADOS_EXT_FUN_COMPILE_FLAGS is used as default.
        """
        return self.__ext_fun_compile_flags


    @property
    def custom_update_filename(self):
        """
        Filename of the custom C function to update solver data and parameters in between solver calls.
        Compare also `AcadosOcpOptions.custom_update_header_filename`.

        This file has to implement the functions
        int custom_update_init_function([model.name]_solver_capsule* capsule);
        int custom_update_function([model.name]_solver_capsule* capsule, double* data, int data_len);
        int custom_update_terminate_function([model.name]_solver_capsule* capsule);


        Default: ''.
        """
        return self.__custom_update_filename


    @property
    def custom_templates(self):
        """
        List of tuples of the form:
        (input_filename, output_filename)

        Custom templates are render in OCP solver generation.

        Default: [].
        """
        return self.__custom_templates


    @property
    def custom_update_header_filename(self):
        """
        Header filename of the custom C function to update solver data and parameters in between solver calls.

        This file has to declare the custom_update functions and look as follows:

        `// Called at the end of solver creation.`

        `// This is allowed to allocate memory and store the pointer to it into capsule->custom_update_memory.`

        `int custom_update_init_function([model.name]_solver_capsule* capsule);`

        `// Custom update function that can be called between solver calls`

        `int custom_update_function([model.name]_solver_capsule* capsule, double* data, int data_len);`

        `// Called just before destroying the solver.`

        `// Responsible to free allocated memory, stored at capsule->custom_update_memory.`

        `int custom_update_terminate_function([model.name]_solver_capsule* capsule);`

        Default: ''.
        """
        return self.__custom_update_header_filename

    @property
    def custom_update_copy(self):
        """
        Boolean;
        If True, the custom update function files are copied into the `code_export_directory`.
        """
        return self.__custom_update_copy


    @property
    def hpipm_mode(self):
        """
        Mode of HPIPM to be used,

        String in ('BALANCE', 'SPEED_ABS', 'SPEED', 'ROBUST').

        Default: 'BALANCE'.

        see https://cdn.syscop.de/publications/Frison2020a.pdf
        and the HPIPM code:
        https://github.com/giaf/hpipm/blob/master/ocp_qp/x_ocp_qp_ipm.c#L69
        """
        return self.__hpipm_mode

    @property
    def hessian_approx(self):
        """Hessian approximation.
        String in ('GAUSS_NEWTON', 'EXACT').
        Default: 'GAUSS_NEWTON'.
        """
        return self.__hessian_approx

    @property
    def integrator_type(self):
        """
        Integrator type.
        String in ('ERK', 'IRK', 'GNSF', 'DISCRETE', 'LIFTED_IRK').
        Default: 'ERK'.
        """
        return self.__integrator_type

    @property
    def nlp_solver_type(self):
        """NLP solver.
        String in ('SQP', 'SQP_RTI', 'DDP').
        Default: 'SQP'.
        """
        return self.__nlp_solver_type

    @property
    def globalization(self):
        """Globalization type.
        String in ('FIXED_STEP', 'MERIT_BACKTRACKING', 'FUNNEL_L1PEN_LINESEARCH').
        Default: 'FIXED_STEP'.

        - FIXED_STEP: performs steps with a given step length, see option globalization_fixed_step_length
        - MERIT_BACKTRACKING: performs a merit function based backtracking line search following Following Leineweber1999, Section "3.5.1 Line Search Globalization"
        - FUNNEL_L1PEN_LINESEARCH: following "A Unified Funnel Restoration SQP Algorithm" by Kiessling et al.
            https://arxiv.org/pdf/2409.09208
            NOTE: preliminary implementation
        """
        return self.__globalization

    @property
    def collocation_type(self):
        """Collocation type: only relevant for implicit integrators
        -- string in {'GAUSS_RADAU_IIA', 'GAUSS_LEGENDRE', 'EXPLICIT_RUNGE_KUTTA'}.

        Default: GAUSS_LEGENDRE
        """
        return self.__collocation_type

    @property
    def regularize_method(self):
        """Regularization method for the Hessian.
        String in ('NO_REGULARIZE', 'MIRROR', 'PROJECT', 'PROJECT_REDUC_HESS', 'CONVEXIFY') or :code:`None`.

        - MIRROR: performs eigenvalue decomposition H = V^T D V and sets D_ii = max(eps, abs(D_ii))
        - PROJECT: performs eigenvalue decomposition H = V^T D V and sets D_ii = max(eps, D_ii)
        - CONVEXIFY: Algorithm 6 from Verschueren2017, https://cdn.syscop.de/publications/Verschueren2017.pdf, does not support nonlinear constraints
        - PROJECT_REDUC_HESS: experimental

        Note: default eps = 1e-4

        Default: 'NO_REGULARIZE'.
        """
        return self.__regularize_method

    @property
    def globalization_fixed_step_length(self):
        """
        Fixed Newton step length, used if globalization == "FIXED_STEP"
        Type: float >= 0.
        Default: 1.0.
        """
        return self.__globalization_fixed_step_length

    @property
    def nlp_solver_step_length(self):
        """
        This option is deprecated and has new name: globalization_fixed_step_length
        """
        print("The option nlp_solver_step_length is deprecated and has new name: globalization_fixed_step_length")
        return self.__globalization_fixed_step_length

    @property
    def nlp_solver_warm_start_first_qp(self):
        """
        Flag indicating whether the first QP in an NLP solve should be warm started.
        Type: int.
        Default: 0.
        """
        return self.__nlp_solver_warm_start_first_qp

    @property
    def levenberg_marquardt(self):
        """
        Factor for LM regularization.
        Type: float >= 0
        Default: 0.0.
        """
        return self.__levenberg_marquardt

    @property
    def sim_method_num_stages(self):
        """
        Number of stages in the integrator.
        Type: int > 0 or ndarray of ints > 0 of shape (N,).
        Default: 4
        """
        return self.__sim_method_num_stages

    @property
    def sim_method_num_steps(self):
        """
        Number of steps in the integrator.
        Type: int > 0 or ndarray of ints > 0 of shape (N,).
        Default: 1
        """
        return self.__sim_method_num_steps

    @property
    def sim_method_newton_iter(self):
        """
        Number of Newton iterations in implicit integrators.
        Type: int > 0
        Default: 3
        """
        return self.__sim_method_newton_iter

    @property
    def sim_method_newton_tol(self):
        """
        Tolerance of Newton system in implicit integrators.
        This option is not implemented for LIFTED_IRK
        Type: float: 0.0 means not used
        Default: 0.0
        """
        return self.__sim_method_newton_tol

    @property
    def sim_method_jac_reuse(self):
        """
        Integer determining if jacobians are reused within integrator or ndarray of ints > 0 of shape (N,).
        0: False (no reuse); 1: True (reuse)
        Default: 0
        """
        return self.__sim_method_jac_reuse

    @property
    def qp_solver_tol_stat(self):
        """
        QP solver stationarity tolerance.
        Default: :code:`None`
        """
        return self.__qp_solver_tol_stat

    @property
    def qp_solver_tol_eq(self):
        """
        QP solver equality tolerance.
        Default: :code:`None`
        """
        return self.__qp_solver_tol_eq

    @property
    def qp_solver_tol_ineq(self):
        """
        QP solver inequality.
        Default: :code:`None`
        """
        return self.__qp_solver_tol_ineq

    @property
    def qp_solver_tol_comp(self):
        """
        QP solver complementarity.
        Default: :code:`None`
        """
        return self.__qp_solver_tol_comp

    @property
    def qp_solver_cond_N(self):
        """QP solver: New horizon after partial condensing.
        Set to N by default -> no condensing."""
        return self.__qp_solver_cond_N

    @property
    def qp_solver_cond_block_size(self):
        """QP solver: list of integers of length qp_solver_cond_N + 1
        Denotes how many blocks of the original OCP are lumped together into one in partial condensing.
        Note that the last entry is the number of blocks that are condensed into the terminal cost of the partially condensed QP.
        Default: None -> compute even block size distribution based on qp_solver_cond_N
        """
        return self.__qp_solver_cond_block_size


    @property
    def qp_solver_warm_start(self):
        """
        QP solver: Warm starting.
        0: no warm start; 1: warm start; 2: hot start.
        Default: 0
        """
        return self.__qp_solver_warm_start

    @property
    def qp_solver_cond_ric_alg(self):
        """
        QP solver: Determines which algorithm is used in HPIPM condensing.
        0: dont factorize hessian in the condensing; 1: factorize.
        Default: 1
        """
        return self.__qp_solver_cond_ric_alg

    @property
    def qp_solver_ric_alg(self):
        """
        QP solver: Determines which algorithm is used in HPIPM OCP QP solver.
        0 classical Riccati, 1 square-root Riccati.

        Note: taken from [HPIPM paper]:

        (a) the classical implementation requires the reduced  (projected) Hessian with respect to the dynamics
            equality constraints to be positive definite, but allows the full-space Hessian to be indefinite)
        (b) the square-root implementation, which in order to reduce the flop count employs the Cholesky
            factorization of the Riccati recursion matrix (P), and therefore requires the full-space Hessian to be positive definite

        [HPIPM paper]: HPIPM: a high-performance quadratic programming framework for model predictive control, Frison and Diehl, 2020
        https://cdn.syscop.de/publications/Frison2020a.pdf

        Default: 1
        """
        return self.__qp_solver_ric_alg

    @property
    def qp_solver_mu0(self):
        """
        For HPIPM QP solver: Initial value for the barrier parameter.
        If 0, the default value according to hpipm_mode is used.

        Default: 0
        """
        return self.__qp_solver_mu0

    @property
    def qp_solver_iter_max(self):
        """
        QP solver: maximum number of iterations.
        Type: int > 0
        Default: 50
        """
        return self.__qp_solver_iter_max


    @property
    def as_rti_iter(self):
        """
        Maximum number of iterations in the advanced-step real-time iteration.
        Default: 1
        """
        return self.__as_rti_iter


    @property
    def as_rti_level(self):
        """
        Level of the advanced-step real-time iteration.

        LEVEL-A: 0
        LEVEL-B: 1
        LEVEL-C: 2
        LEVEL-D: 3
        STANDARD_RTI: 4

        Default: 4
        """
        return self.__as_rti_level

    @property
    def with_adaptive_levenberg_marquardt(self):
        """
        So far: Only relevant for DDP
        Flag indicating whether adaptive levenberg marquardt should be used.
        This is useful for NLS or LS problem where the residual goes to zero since
        quadratic local convergence can be achieved.
        type: bool
        """
        return self.__with_adaptive_levenberg_marquardt

    @property
    def adaptive_levenberg_marquardt_lam(self):
        """
        So far: Only relevant for DDP
        Flag defining the value of lambda in the adaptive levenberg_marquardt.
        Must be > 1.
        type: float
        """
        return self.__adaptive_levenberg_marquardt_lam

    @property
    def adaptive_levenberg_marquardt_mu_min(self):
        """
        So far: Only relevant for DDP
        Flag defining the value of mu_min in the adaptive levenberg_marquardt.
        Must be > 0.
        type: float
        """
        return self.__adaptive_levenberg_marquardt_mu_min

    @property
    def adaptive_levenberg_marquardt_mu0(self):
        """
        So far: Only relevant for DDP
        Flag defining the value of mu0 in the adaptive levenberg_marquardt.
        Must be > 0.
        type: float
        """
        return self.__adaptive_levenberg_marquardt_mu0

    @property
    def log_primal_step_norm(self,):
        """
        Flag indicating whether the max norm of the primal steps should be logged.
        This is implemented only for solver type `SQP`.
        Default: False
        """
        return self.__log_primal_step_norm

    @property
    def store_iterates(self,):
        """
        Flag indicating whether the intermediate primal-dual iterates should be stored.
        This is implemented only for solver types `SQP` and `DDP`.
        Default: False
        """
        return self.__store_iterates

    @property
    def timeout_max_time(self,):
        """
        Maximum time before solver timeout. If 0, there is no timeout.
        A timeout is triggered if the condition
        `current_time_tot + predicted_per_iteration_time > timeout_max_time`
        is satisfied at the end of an SQP iteration.
        The value of `predicted_per_iteration_time` is estimated using `timeout_heuristic`.
        Currently implemented for SQP only.
        Default: 0.
        """
        return self.__timeout_max_time

    @property
    def timeout_heuristic(self,):
        """
        Heuristic to be used for predicting the runtime of the next SQP iteration, cf. `timeout_max_time`.
        Possible values are "MAX_CALL", "MAX_OVERALL", "LAST", "AVERAGE", "ZERO".
        MAX_CALL: Use the maximum time per iteration for the current solver call as estimate.
        MAX_OVERALL: Use the maximum time per iteration over all solver calls as estimate.
        LAST: Use the time required by the last iteration as estimate.
        AVERAGE: Use an exponential moving average of the previous per iteration times as estimate (weight is currently fixed at 0.5).
        ZERO: Use 0 as estimate.
        Currently implemented for SQP only.
        Default: ZERO.
        """
        return self.__timeout_heuristic

    @property
    def tol(self):
        """
        NLP solver tolerance. Sets or gets the max of :py:attr:`nlp_solver_tol_eq`,
        :py:attr:`nlp_solver_tol_ineq`, :py:attr:`nlp_solver_tol_comp`
        and :py:attr:`nlp_solver_tol_stat`.
        """
        return max([self.__nlp_solver_tol_eq, self.__nlp_solver_tol_ineq,\
                    self.__nlp_solver_tol_comp, self.__nlp_solver_tol_stat])

    @property
    def qp_tol(self):
        """
        QP solver tolerance.
        Sets all of the following at once or gets the max of
        :py:attr:`qp_solver_tol_eq`, :py:attr:`qp_solver_tol_ineq`,
        :py:attr:`qp_solver_tol_comp` and
        :py:attr:`qp_solver_tol_stat`.
        """
        return max([self.__qp_solver_tol_eq, self.__qp_solver_tol_ineq,\
                    self.__qp_solver_tol_comp, self.__qp_solver_tol_stat])

    @property
    def nlp_solver_tol_stat(self):
        """
        NLP solver stationarity tolerance.
        Type: float > 0
        Default: 1e-6
        """
        return self.__nlp_solver_tol_stat

    @property
    def nlp_solver_tol_eq(self):
        """NLP solver equality tolerance"""
        return self.__nlp_solver_tol_eq

    @property
    def nlp_solver_tol_min_step_norm(self):
        """
        NLP solver tolerance for minimal step norm. Solver terminates if
        step norm is below given value. If value is 0.0, then the solver does not
        test for the small step.

        Type: float
        Default: None

        If None:
        in case of FUNNEL_L1PEN_LINESEARCH: 1e-12
        otherwise: 0.0
         """
        return self.__nlp_solver_tol_min_step_norm

    @property
    def globalization_alpha_min(self):
        """Minimal step size for globalization.

        default: None.

        If None is given:
        - in case of FUNNEL_L1PEN_LINESEARCH, value is set to 1e-17.
        - in case of MERIT_BACKTRACKING, value is set to 0.05.
        """
        return self.__globalization_alpha_min

    @property
    def alpha_min(self):
        """
        The option alpha_min is deprecated and has new name: globalization_alpha_min
        """
        print("The option alpha_min is deprecated and has new name: globalization_alpha_min")
        return self.globalization_alpha_min

    @property
    def reg_epsilon(self):
        """Epsilon for regularization, used if regularize_method in ['PROJECT', 'MIRROR', 'CONVEXIFY']"""
        return self.__reg_epsilon

    @property
    def globalization_alpha_reduction(self):
        """Step size reduction factor for globalization MERIT_BACKTRACKING,

        Type: float
        Default: None.

        If None is given:
        - in case of FUNNEL_L1PEN_LINESEARCH, value is set to 0.5.
        - in case of MERIT_BACKTRACKING, value is set to 0.7.
        default: 0.7.
        """
        return self.__globalization_alpha_reduction

    @property
    def alpha_reduction(self):
        """
        The option alpha_reduction is deprecated and has new name: globalization_alpha_reduction
        """
        print("The option alpha_reduction is deprecated and has new name: globalization_alpha_reduction")
        return self.globalization_alpha_reduction

    @property
    def globalization_line_search_use_sufficient_descent(self):
        """
        Determines if sufficient descent (Armijo) condition is used in line search.
        Type: int; 0 or 1;
        default: 0.
        """
        return self.__globalization_line_search_use_sufficient_descent

    @property
    def line_search_use_sufficient_descent(self):
        """
        The option line_search_use_sufficient_descent is deprecated and has new name: globalization_line_search_use_sufficient_descent
        """
        print("The option line_search_use_sufficient_descent is deprecated and has new name: globalization_line_search_use_sufficient_descent")
        return self.globalization_line_search_use_sufficient_descent

    @property
    def globalization_eps_sufficient_descent(self):
        """
        Factor for sufficient descent (Armijo) conditon, see also globalization_line_search_use_sufficient_descent.

        Type: float,
        Default: None.

        If None is given:
        - in case of FUNNEL_L1PEN_LINESEARCH, value is set to 1e-6.
        - in case of MERIT_BACKTRACKING, value is set to 1e-4.
        """
        return self.__globalization_eps_sufficient_descent

    @property
    def eps_sufficient_descent(self):
        """
        The option eps_sufficient_descent is deprecated and has new name: globalization_line_search_use_sufficient_descent
        """
        print("The option eps_sufficient_descent is deprecated and has new name: globalization_line_search_use_sufficient_descent")
        return self.globalization_line_search_use_sufficient_descent

    @property
    def globalization_use_SOC(self):
        """
        Determines if second order correction (SOC) is done when using MERIT_BACKTRACKING.
        SOC is done if preliminary line search does not return full step.
        Type: int; 0 or 1;
        default: 0.
        """
        return self.__globalization_use_SOC

    @property
    def globalization_full_step_dual(self):
        """
        Determines if dual variables are updated with full steps (alpha=1.0) when primal variables are updated with smaller step.

        Type: int; 0 or 1;
        default for funnel globalization: 1
        default else: 0.
        """
        return self.__globalization_full_step_dual

    @property
    def full_step_dual(self):
        """
        The option full_step_dual is deprecated and has new name: globalization_full_step_dual
        """
        print("The option full_step_dual is deprecated and has new name: globalization_full_step_dual")
        return self.globalization_full_step_dual

    @property
    def globalization_funnel_init_increase_factor(self):
        """
        Increase factor for initialization of funnel width.
        Initial funnel is max(globalization_funnel_init_upper_bound, globalization_funnel_init_increase_factor * initial_infeasibility)

        Type: float
        Default: 15.0
        """
        return self.__globalization_funnel_init_increase_factor

    @property
    def globalization_funnel_init_upper_bound(self):
        """
        Initial upper bound for funnel width.
        Initial funnel is max(globalization_funnel_init_upper_bound, globalization_funnel_init_increase_factor * initial_infeasibility)

        Type: float
        Default: 1.0
        """
        return self.__globalization_funnel_init_upper_bound

    @property
    def globalization_funnel_sufficient_decrease_factor(self):
        """
        Sufficient decrease factor for infeasibility in h iteration:
        trial_infeasibility <= kappa * globalization_funnel_width

        Type: float
        Default: 0.9
        """
        return self.__globalization_funnel_sufficient_decrease_factor

    @property
    def globalization_funnel_kappa(self):
        """
        Interpolation factor for convex combination in funnel decrease function.

        Type: float
        Default: 0.9
        """
        return self.__globalization_funnel_kappa

    @property
    def globalization_funnel_fraction_switching_condition(self):
        """
        Multiplication factor in switching condition.

        Type: float
        Default: 1e-3
        """
        return self.__globalization_funnel_fraction_switching_condition

    @property
    def eval_residual_at_max_iter(self):
        """
        Determines, if the problem data is again evaluated after the last iteration
        has been performed.
        If True, the residuals are evaluated at the last iterate and convergence will be checked.
        If False, after the last iteration, the solver will terminate with max iterations
        status, although the final iterate might fulfill the convergence criteria.

        Type: bool
        Default: None

        If None is given:
        if globalization == FUNNEL_L1PEN_LINESEARCH: true
        else: false
        """
        return self.__eval_residual_at_max_iter

    @property
    def globalization_funnel_initial_penalty_parameter(self):
        """
        Initialization.

        Type: float
        Default: 1.0
        """
        return self.__globalization_funnel_initial_penalty_parameter

    @property
    def nlp_solver_tol_ineq(self):
        """NLP solver inequality tolerance"""
        return self.__nlp_solver_tol_ineq

    @property
    def nlp_solver_ext_qp_res(self):
        """Determines if residuals of QP are computed externally within NLP solver (for debugging)

        Type: int; 0 or 1;
        Default: 0.
        """
        return self.__nlp_solver_ext_qp_res

    @property
    def rti_log_residuals(self):
        """Determines if residuals are computed and logged within RTI / AS-RTI iterations (for debugging).

        Type: int; 0 or 1;
        Default: 0.
        """
        return self.__rti_log_residuals

    @property
    def rti_log_only_available_residuals(self):
        """
        Relevant if rti_log_residuals is set to 1.
        If rti_log_only_available_residuals is set to 1, only residuals that do not require additional function evaluations are logged.

        Type: int; 0 or 1;
        Default: 0.
        """
        return self.__rti_log_only_available_residuals

    @property
    def nlp_solver_tol_comp(self):
        """NLP solver complementarity tolerance"""
        return self.__nlp_solver_tol_comp

    @property
    def nlp_solver_max_iter(self):
        """
        NLP solver maximum number of iterations.
        Type: int >= 0
        Default: 100
        """
        return self.__nlp_solver_max_iter

    @property
    def time_steps(self):
        """
        Vector of length `N_horizon` containing the time steps between the shooting nodes.
        If `None` set automatically to uniform discretization using :py:attr:`N_horizon` and :py:attr:`tf`.
        For nonuniform discretization: Either provide shooting_nodes or time_steps.
        Default: :code:`None`
        """
        return self.__time_steps

    @property
    def shooting_nodes(self):
        """
        Vector of length `N_horizon + 1` containing the shooting nodes.
        If `None` set automatically to uniform discretization using :py:attr:`N_horizon` and :py:attr:`tf`.
        For nonuniform discretization: Either provide shooting_nodes or time_steps.
        Default: :code:`None`
        """
        return self.__shooting_nodes

    @property
    def cost_scaling(self):
        """
        Vector with cost scaling factors of length `N_horizon` + 1.
        If `None` set automatically to [`time_steps`, 1.0].
        Default: :code:`None`
        """
        return self.__cost_scaling

    @property
    def tf(self):
        """
        Prediction horizon
        Type: float > 0
        Default: :code:`None`
        """
        return self.__tf

    @property
    def N_horizon(self):
        """
        Number of shooting intervals.
        Type: int > 0
        Default: :code:`None`
        """
        return self.__N_horizon

    @property
    def Tsim(self):
        """
        Time horizon for one integrator step. Automatically calculated as first time step if not provided.
        Default: :code:`None`
        """
        return self.__Tsim

    @property
    def print_level(self):
        """
        Verbosity of printing.

        Type: int >= 0
        Default: 0

        Level 1: print iteration log
        Level 2: print high level debug output in funnel globalization
        Level 3: print more detailed debug output in funnel, including objective values and infeasibilities
        Level 4: print QP inputs and outputs. Please specify with max_iter how many QPs should be printed
        """
        return self.__print_level

    @property
    def model_external_shared_lib_dir(self):
        """Path to the .so lib"""
        return self.__model_external_shared_lib_dir

    @property
    def model_external_shared_lib_name(self):
        """Name of the .so lib"""
        return self.__model_external_shared_lib_name

    @property
    def exact_hess_constr(self):
        """
        Used in case of hessian_approx == 'EXACT'.\n
        Can be used to turn off exact hessian contributions from the constraints module.
        """
        return self.__exact_hess_constr

    @property
    def exact_hess_cost(self):
        """
        Used in case of hessian_approx == 'EXACT'.\n
        Can be used to turn off exact hessian contributions from the cost module.
        """
        return self.__exact_hess_cost

    @property
    def exact_hess_dyn(self):
        """
        Used in case of hessian_approx == 'EXACT'.\n
        Can be used to turn off exact hessian contributions from the dynamics module.
        """
        return self.__exact_hess_dyn

    @property
    def fixed_hess(self):
        """
        Indicates if the hessian is fixed (1) or not (0).\n
        If fixed, the hessian is computed only once and not updated in the SQP loop.
        This can safely be set to 1 if there are no slacked constraints, cost module is 'LINEAR_LS' with Gauss-Newton Hessian,
        and the weighting matrix 'W' is not updated after solver creation.
        Default: 0
        """
        return self.__fixed_hess

    @property
    def ext_cost_num_hess(self):
        """
        Determines if custom hessian approximation for cost contribution is used (> 0).\n
        Or if hessian contribution is evaluated exactly using CasADi external function (=0 - default).
        """
        return self.__ext_cost_num_hess

    @property
    def cost_discretization(self):
        """
        Cost discretization: string in {'EULER', 'INTEGRATOR'}.
        Default: 'EULER'
        'EULER': cost is evaluated at shooting nodes
        'INTEGRATOR': cost is integrated over the shooting intervals - only supported for IRK integrator
        """
        return self.__cost_discretization

    @property
    def with_solution_sens_wrt_params(self):
        """
        Flag indicating whether solution sensitivities wrt. parameters can be computed.
        """
        return self.__with_solution_sens_wrt_params

    @property
    def with_value_sens_wrt_params(self):
        """
        Flag indicating whether value function sensitivities wrt. parameters can be computed.
        """
        return self.__with_value_sens_wrt_params

    @property
    def num_threads_in_batch_solve(self):
        """
        Integer indicating how many threads should be used within the batch solve.
        If more than one thread should be used, the solver is compiled with openmp.
        Default: 1.
        """
        return self.__num_threads_in_batch_solve


    @qp_solver.setter
    def qp_solver(self, qp_solver):
        qp_solvers = ('PARTIAL_CONDENSING_HPIPM', \
                'FULL_CONDENSING_QPOASES', 'FULL_CONDENSING_HPIPM', \
                'PARTIAL_CONDENSING_QPDUNES', 'PARTIAL_CONDENSING_OSQP', \
                'FULL_CONDENSING_DAQP')
        if qp_solver in qp_solvers:
            self.__qp_solver = qp_solver
        else:
            raise Exception('Invalid qp_solver value. Possible values are:\n\n' \
                    + ',\n'.join(qp_solvers) + '.\n\nYou have: ' + qp_solver + '.\n\n')


    @regularize_method.setter
    def regularize_method(self, regularize_method):
        regularize_methods = ('NO_REGULARIZE', 'MIRROR', 'PROJECT', \
                                'PROJECT_REDUC_HESS', 'CONVEXIFY')
        if regularize_method in regularize_methods:
            self.__regularize_method = regularize_method
        else:
            raise Exception('Invalid regularize_method value. Possible values are:\n\n' \
                    + ',\n'.join(regularize_methods) + '.\n\nYou have: ' + regularize_method + '.\n\n')

    @collocation_type.setter
    def collocation_type(self, collocation_type):
        if collocation_type in COLLOCATION_TYPES:
            self.__collocation_type = collocation_type
        else:
            raise Exception('Invalid collocation_type value. Possible values are:\n\n' \
                    + ',\n'.join(COLLOCATION_TYPES) + '.\n\nYou have: ' + collocation_type + '.\n\n')

    @hpipm_mode.setter
    def hpipm_mode(self, hpipm_mode):
        hpipm_modes = ('BALANCE', 'SPEED_ABS', 'SPEED', 'ROBUST')
        if hpipm_mode in hpipm_modes:
            self.__hpipm_mode = hpipm_mode
        else:
            raise Exception('Invalid hpipm_mode value. Possible values are:\n\n' \
                    + ',\n'.join(hpipm_modes) + '.\n\nYou have: ' + hpipm_mode + '.\n\n')

    @with_solution_sens_wrt_params.setter
    def with_solution_sens_wrt_params(self, with_solution_sens_wrt_params):
        if isinstance(with_solution_sens_wrt_params, bool):
            self.__with_solution_sens_wrt_params = with_solution_sens_wrt_params
        else:
            raise Exception('Invalid with_solution_sens_wrt_params value. Expected bool.')

    @with_value_sens_wrt_params.setter
    def with_value_sens_wrt_params(self, with_value_sens_wrt_params):
        if isinstance(with_value_sens_wrt_params, bool):
            self.__with_value_sens_wrt_params = with_value_sens_wrt_params
        else:
            raise Exception('Invalid with_value_sens_wrt_params value. Expected bool.')

    @ext_fun_compile_flags.setter
    def ext_fun_compile_flags(self, ext_fun_compile_flags):
        if isinstance(ext_fun_compile_flags, str):
            self.__ext_fun_compile_flags = ext_fun_compile_flags
        else:
            raise Exception('Invalid ext_fun_compile_flags, expected a string.\n')


    @custom_update_filename.setter
    def custom_update_filename(self, custom_update_filename):
        if isinstance(custom_update_filename, str):
            self.__custom_update_filename = custom_update_filename
        else:
            raise Exception('Invalid custom_update_filename, expected a string.\n')

    @custom_templates.setter
    def custom_templates(self, custom_templates):
        if not isinstance(custom_templates, list):
            raise Exception('Invalid custom_templates, expected a list.\n')
        for tup in custom_templates:
            if not isinstance(tup, tuple):
                raise Exception('Invalid custom_templates, shoubld be list of tuples.\n')
            for s in tup:
                if not isinstance(s, str):
                    raise Exception('Invalid custom_templates, shoubld be list of tuples of strings.\n')
        self.__custom_templates = custom_templates

    @custom_update_header_filename.setter
    def custom_update_header_filename(self, custom_update_header_filename):
        if isinstance(custom_update_header_filename, str):
            self.__custom_update_header_filename = custom_update_header_filename
        else:
            raise Exception('Invalid custom_update_header_filename, expected a string.\n')

    @custom_update_copy.setter
    def custom_update_copy(self, custom_update_copy):
        if isinstance(custom_update_copy, bool):
            self.__custom_update_copy = custom_update_copy
        else:
            raise Exception('Invalid custom_update_copy, expected a bool.\n')

    @hessian_approx.setter
    def hessian_approx(self, hessian_approx):
        hessian_approxs = ('GAUSS_NEWTON', 'EXACT')
        if hessian_approx in hessian_approxs:
            self.__hessian_approx = hessian_approx
        else:
            raise Exception('Invalid hessian_approx value. Possible values are:\n\n' \
                    + ',\n'.join(hessian_approxs) + '.\n\nYou have: ' + hessian_approx + '.\n\n')

    @integrator_type.setter
    def integrator_type(self, integrator_type):
        if integrator_type in INTEGRATOR_TYPES:
            self.__integrator_type = integrator_type
        else:
            raise Exception('Invalid integrator_type value. Possible values are:\n\n' \
                    + ',\n'.join(INTEGRATOR_TYPES) + '.\n\nYou have: ' + integrator_type + '.\n\n')

    @tf.setter
    def tf(self, tf):
        self.__tf = tf

    @N_horizon.setter
    def N_horizon(self, N_horizon):
        if isinstance(N_horizon, int) and N_horizon > 0:
            self.__N_horizon = N_horizon
        else:
            raise Exception('Invalid N_horizon value, expected positive integer.')

    @time_steps.setter
    def time_steps(self, time_steps):
        time_steps = check_if_nparray_and_flatten(time_steps, "time_steps")
        self.__time_steps = time_steps

    @shooting_nodes.setter
    def shooting_nodes(self, shooting_nodes):
        shooting_nodes = check_if_nparray_and_flatten(shooting_nodes, "shooting_nodes")
        self.__shooting_nodes = shooting_nodes

    @cost_scaling.setter
    def cost_scaling(self, cost_scaling):
        cost_scaling = check_if_nparray_and_flatten(cost_scaling, "cost_scaling")
        self.__cost_scaling = cost_scaling

    @Tsim.setter
    def Tsim(self, Tsim):
        self.__Tsim = Tsim

    @globalization.setter
    def globalization(self, globalization):
        globalization_types = ('FUNNEL_L1PEN_LINESEARCH', 'MERIT_BACKTRACKING', 'FIXED_STEP')
        if globalization in globalization_types:
            self.__globalization = globalization
        else:
            raise Exception('Invalid globalization value. Possible values are:\n\n' \
                    + ',\n'.join(globalization_types) + '.\n\nYou have: ' + globalization + '.\n\n')

    @reg_epsilon.setter
    def reg_epsilon(self, reg_epsilon):
        self.__reg_epsilon = reg_epsilon

    @globalization_alpha_min.setter
    def globalization_alpha_min(self, globalization_alpha_min):
        self.__globalization_alpha_min = globalization_alpha_min

    @alpha_min.setter
    def alpha_min(self, alpha_min):
        print("This option is deprecated and has new name: globalization_alpha_min")
        self.globalization_alpha_min = alpha_min

    @globalization_alpha_reduction.setter
    def globalization_alpha_reduction(self, globalization_alpha_reduction):
        self.__globalization_alpha_reduction = globalization_alpha_reduction

    @alpha_reduction.setter
    def alpha_reduction(self, globalization_alpha_reduction):
        print("This option is deprecated and has new name: globalization_alpha_reduction")
        self.globalization_alpha_reduction = globalization_alpha_reduction

    @globalization_line_search_use_sufficient_descent.setter
    def globalization_line_search_use_sufficient_descent(self, globalization_line_search_use_sufficient_descent):
        if globalization_line_search_use_sufficient_descent in [0, 1]:
            self.__globalization_line_search_use_sufficient_descent = globalization_line_search_use_sufficient_descent
        else:
            raise Exception(f'Invalid value for globalization_line_search_use_sufficient_descent. Possible values are 0, 1, got {globalization_line_search_use_sufficient_descent}')

    @line_search_use_sufficient_descent.setter
    def line_search_use_sufficient_descent(self, globalization_line_search_use_sufficient_descent):
        print("This option is deprecated and has new name: globalization_line_search_use_sufficient_descent")
        if globalization_line_search_use_sufficient_descent in [0, 1]:
            self.__globalization_line_search_use_sufficient_descent = globalization_line_search_use_sufficient_descent
        else:
            raise Exception(f'Invalid value for globalization_line_search_use_sufficient_descent. Possible values are 0, 1, got {globalization_line_search_use_sufficient_descent}')

    @globalization_use_SOC.setter
    def globalization_use_SOC(self, globalization_use_SOC):
        if globalization_use_SOC in [0, 1]:
            self.__globalization_use_SOC = globalization_use_SOC
        else:
            raise Exception(f'Invalid value for globalization_use_SOC. Possible values are 0, 1, got {globalization_use_SOC}')

    @globalization_full_step_dual.setter
    def globalization_full_step_dual(self, globalization_full_step_dual):
        if globalization_full_step_dual in [0, 1]:
            self.__globalization_full_step_dual = globalization_full_step_dual
        else:
            raise Exception(f'Invalid value for globalization_full_step_dual. Possible values are 0, 1, got {globalization_full_step_dual}')

    @full_step_dual.setter
    def full_step_dual(self, globalization_full_step_dual):
        print("This option is deprecated and has new name: globalization_full_step_dual")
        self.globalization_full_step_dual = globalization_full_step_dual


    @globalization_funnel_init_increase_factor.setter
    def globalization_funnel_init_increase_factor(self, globalization_funnel_init_increase_factor):
        if globalization_funnel_init_increase_factor > 1.0:
            self.__globalization_funnel_init_increase_factor = globalization_funnel_init_increase_factor
        else:
            raise Exception(f'Invalid value for globalization_funnel_init_increase_factor. Should be > 1, got {globalization_funnel_init_increase_factor}')

    @globalization_funnel_init_upper_bound.setter
    def globalization_funnel_init_upper_bound(self, globalization_funnel_init_upper_bound):
        if globalization_funnel_init_upper_bound > 0.0:
            self.__globalization_funnel_init_upper_bound = globalization_funnel_init_upper_bound
        else:
             raise Exception(f'Invalid value for globalization_funnel_init_upper_bound. Should be > 0, got {globalization_funnel_init_upper_bound}')

    @globalization_funnel_sufficient_decrease_factor.setter
    def globalization_funnel_sufficient_decrease_factor(self, globalization_funnel_sufficient_decrease_factor):
        if globalization_funnel_sufficient_decrease_factor > 0.0 and globalization_funnel_sufficient_decrease_factor < 1.0:
            self.__globalization_funnel_sufficient_decrease_factor = globalization_funnel_sufficient_decrease_factor
        else:
            raise Exception(f'Invalid value for globalization_funnel_sufficient_decrease_factor. Should be in (0,1), got {globalization_funnel_sufficient_decrease_factor}')

    @globalization_funnel_kappa.setter
    def globalization_funnel_kappa(self, globalization_funnel_kappa):
        if globalization_funnel_kappa > 0.0 and globalization_funnel_kappa < 1.0:
            self.__globalization_funnel_kappa = globalization_funnel_kappa
        else:
            raise Exception(f'Invalid value for globalization_funnel_kappa. Should be in (0,1), got {globalization_funnel_kappa}')

    @globalization_funnel_fraction_switching_condition.setter
    def globalization_funnel_fraction_switching_condition(self, globalization_funnel_fraction_switching_condition):
        if globalization_funnel_fraction_switching_condition > 0.0 and globalization_funnel_fraction_switching_condition < 1.0:
            self.__globalization_funnel_fraction_switching_condition = globalization_funnel_fraction_switching_condition
        else:
            raise Exception(f'Invalid value for globalization_funnel_fraction_switching_condition. Should be in (0,1), got {globalization_funnel_fraction_switching_condition}')

    @globalization_funnel_initial_penalty_parameter.setter
    def globalization_funnel_initial_penalty_parameter(self, globalization_funnel_initial_penalty_parameter):
        if globalization_funnel_initial_penalty_parameter >= 0.0 and globalization_funnel_initial_penalty_parameter <= 1.0:
            self.__globalization_funnel_initial_penalty_parameter = globalization_funnel_initial_penalty_parameter
        else:
            raise Exception(f'Invalid value for globalization_funnel_initial_penalty_parameter. Should be in [0,1], got {globalization_funnel_initial_penalty_parameter}')

    @eval_residual_at_max_iter.setter
    def eval_residual_at_max_iter(self, eval_residual_at_max_iter):
        if isinstance(eval_residual_at_max_iter, bool):
            self.__eval_residual_at_max_iter = eval_residual_at_max_iter
        else:
            raise Exception(f'Invalid datatype for eval_residual_at_max_iter. Should be bool, got {type(eval_residual_at_max_iter)}')

    @globalization_eps_sufficient_descent.setter
    def globalization_eps_sufficient_descent(self, globalization_eps_sufficient_descent):
        if isinstance(globalization_eps_sufficient_descent, float) and globalization_eps_sufficient_descent > 0:
            self.__globalization_eps_sufficient_descent = globalization_eps_sufficient_descent
        else:
            raise Exception('Invalid globalization_eps_sufficient_descent value. globalization_eps_sufficient_descent must be a positive float.')

    @eps_sufficient_descent.setter
    def eps_sufficient_descent(self, globalization_eps_sufficient_descent):
        print("This option is deprecated and has new name: globalization_eps_sufficient_descent")
        self.globalization_eps_sufficient_descent = globalization_eps_sufficient_descent

    @sim_method_num_stages.setter
    def sim_method_num_stages(self, sim_method_num_stages):

        # if isinstance(sim_method_num_stages, int):
        #     self.__sim_method_num_stages = sim_method_num_stages
        # else:
        #     raise Exception('Invalid sim_method_num_stages value. sim_method_num_stages must be an integer.')

        self.__sim_method_num_stages = sim_method_num_stages

    @sim_method_num_steps.setter
    def sim_method_num_steps(self, sim_method_num_steps):

        # if isinstance(sim_method_num_steps, int):
        #     self.__sim_method_num_steps = sim_method_num_steps
        # else:
        #     raise Exception('Invalid sim_method_num_steps value. sim_method_num_steps must be an integer.')
        self.__sim_method_num_steps = sim_method_num_steps


    @sim_method_newton_iter.setter
    def sim_method_newton_iter(self, sim_method_newton_iter):

        if isinstance(sim_method_newton_iter, int):
            self.__sim_method_newton_iter = sim_method_newton_iter
        else:
            raise Exception('Invalid sim_method_newton_iter value. sim_method_newton_iter must be an integer.')

    @sim_method_newton_tol.setter
    def sim_method_newton_tol(self, sim_method_newton_tol):
        if isinstance(sim_method_newton_tol, float) and sim_method_newton_tol > 0:
            self.__sim_method_newton_tol = sim_method_newton_tol
        else:
            raise Exception('Invalid sim_method_newton_tol value. sim_method_newton_tol must be a positive float.')

    @sim_method_jac_reuse.setter
    def sim_method_jac_reuse(self, sim_method_jac_reuse):
        self.__sim_method_jac_reuse = sim_method_jac_reuse

    @nlp_solver_type.setter
    def nlp_solver_type(self, nlp_solver_type):
        nlp_solver_types = ('SQP', 'SQP_RTI', 'DDP')
        if nlp_solver_type in nlp_solver_types:
            self.__nlp_solver_type = nlp_solver_type
        else:
            raise Exception('Invalid nlp_solver_type value. Possible values are:\n\n' \
                    + ',\n'.join(nlp_solver_types) + '.\n\nYou have: ' + nlp_solver_type + '.\n\n')

    @cost_discretization.setter
    def cost_discretization(self, cost_discretization):
        if cost_discretization in COST_DISCRETIZATION_TYPES:
            self.__cost_discretization = cost_discretization
        else:
            raise Exception('Invalid cost_discretization value. Possible values are:\n\n' \
                    + ',\n'.join(COST_DISCRETIZATION_TYPES) + '.\n\nYou have: ' + cost_discretization + '.')

    @globalization_fixed_step_length.setter
    def globalization_fixed_step_length(self, globalization_fixed_step_length):
        if isinstance(globalization_fixed_step_length, float) and globalization_fixed_step_length >= 0.:
            self.__globalization_fixed_step_length = globalization_fixed_step_length
        else:
            raise Exception('Invalid globalization_fixed_step_length value. globalization_fixed_step_length must be a positive float.')

    @nlp_solver_step_length.setter
    def nlp_solver_step_length(self, nlp_solver_step_length):
        print("The option nlp_solver_step_length is deprecated and has new name: globalization_fixed_step_length")
        self.globalization_fixed_step_length = nlp_solver_step_length

    @nlp_solver_warm_start_first_qp.setter
    def nlp_solver_warm_start_first_qp(self, nlp_solver_warm_start_first_qp):
        if isinstance(nlp_solver_warm_start_first_qp, bool):
            self.__nlp_solver_warm_start_first_qp = nlp_solver_warm_start_first_qp
        else:
            raise Exception('Invalid nlp_solver_warm_start_first_qp value. Expected bool.')

    @levenberg_marquardt.setter
    def levenberg_marquardt(self, levenberg_marquardt):
        if isinstance(levenberg_marquardt, float) and levenberg_marquardt >= 0:
            self.__levenberg_marquardt = levenberg_marquardt
        else:
            raise Exception('Invalid levenberg_marquardt value. levenberg_marquardt must be a positive float.')

    @qp_solver_mu0.setter
    def qp_solver_mu0(self, qp_solver_mu0):
        if isinstance(qp_solver_mu0, float) and qp_solver_mu0 >= 0:
            self.__qp_solver_mu0 = qp_solver_mu0
        else:
            raise Exception('Invalid qp_solver_mu0 value. qp_solver_mu0 must be a positive float.')

    @qp_solver_iter_max.setter
    def qp_solver_iter_max(self, qp_solver_iter_max):
        if isinstance(qp_solver_iter_max, int) and qp_solver_iter_max > 0:
            self.__qp_solver_iter_max = qp_solver_iter_max
        else:
            raise Exception('Invalid qp_solver_iter_max value. qp_solver_iter_max must be a positive int.')

    @with_adaptive_levenberg_marquardt.setter
    def with_adaptive_levenberg_marquardt(self, with_adaptive_levenberg_marquardt):
        if isinstance(with_adaptive_levenberg_marquardt, bool):
            self.__with_adaptive_levenberg_marquardt = with_adaptive_levenberg_marquardt
        else:
            raise Exception('Invalid with_adaptive_levenberg_marquardt value. Expected bool.')

    @adaptive_levenberg_marquardt_lam.setter
    def adaptive_levenberg_marquardt_lam(self, adaptive_levenberg_marquardt_lam):
        if isinstance(adaptive_levenberg_marquardt_lam, float) and adaptive_levenberg_marquardt_lam >= 1.0:
            self.__adaptive_levenberg_marquardt_lam = adaptive_levenberg_marquardt_lam
        else:
            raise Exception('Invalid adaptive_levenberg_marquardt_lam value. adaptive_levenberg_marquardt_lam must be a float greater 1.0.')

    @adaptive_levenberg_marquardt_mu_min.setter
    def adaptive_levenberg_marquardt_mu_min(self, adaptive_levenberg_marquardt_mu_min):
        if isinstance(adaptive_levenberg_marquardt_mu_min, float) and adaptive_levenberg_marquardt_mu_min >= 0.0:
            self.__adaptive_levenberg_marquardt_mu_min = adaptive_levenberg_marquardt_mu_min
        else:
            raise Exception('Invalid adaptive_levenberg_marquardt_mu_min value. adaptive_levenberg_marquardt_mu_min must be a positive float.')

    @adaptive_levenberg_marquardt_mu0.setter
    def adaptive_levenberg_marquardt_mu0(self, adaptive_levenberg_marquardt_mu0):
        if isinstance(adaptive_levenberg_marquardt_mu0, float) and adaptive_levenberg_marquardt_mu0 >= 0.0:
            self.__adaptive_levenberg_marquardt_mu0 = adaptive_levenberg_marquardt_mu0
        else:
            raise Exception('Invalid adaptive_levenberg_marquardt_mu0 value. adaptive_levenberg_marquardt_mu0 must be a positive float.')

    @log_primal_step_norm.setter
    def log_primal_step_norm(self, val):
        if isinstance(val, bool):
            self.__log_primal_step_norm = val
        else:
            raise Exception('Invalid log_primal_step_norm value. Expected bool.')

    @store_iterates.setter
    def store_iterates(self, val):
        if isinstance(val, bool):
            self.__store_iterates = val
        else:
            raise Exception('Invalid store_iterates value. Expected bool.')

    @timeout_max_time.setter
    def timeout_max_time(self, val):
        if isinstance(val, float) and val >= 0:
            self.__timeout_max_time = val
        else:
            raise Exception('Invalid timeout_max_time value. Expected nonnegative float.')

    @timeout_heuristic.setter
    def timeout_heuristic(self, val):
        if val in ["MAX_CALL", "MAX_OVERALL", "LAST", "AVERAGE", "ZERO"]:
            self.__timeout_heuristic = val
        else:
            raise Exception('Invalid timeout_heuristic value. Expected value in ["MAX_CALL", "MAX_OVERALL", "LAST", "AVERAGE", "ZERO"].')

    @as_rti_iter.setter
    def as_rti_iter(self, as_rti_iter):
        if isinstance(as_rti_iter, int) and as_rti_iter >= 0:
            self.__as_rti_iter = as_rti_iter
        else:
            raise Exception('Invalid as_rti_iter value. as_rti_iter must be a nonnegative int.')

    @as_rti_level.setter
    def as_rti_level(self, as_rti_level):
        if as_rti_level in [0, 1, 2, 3, 4]:
            self.__as_rti_level = as_rti_level
        else:
            raise Exception('Invalid as_rti_level value must be in [0, 1, 2, 3, 4].')


    @qp_solver_ric_alg.setter
    def qp_solver_ric_alg(self, qp_solver_ric_alg):
        if qp_solver_ric_alg in [0, 1]:
            self.__qp_solver_ric_alg = qp_solver_ric_alg
        else:
            raise Exception(f'Invalid qp_solver_ric_alg value. qp_solver_ric_alg must be in [0, 1], got {qp_solver_ric_alg}.')

    @qp_solver_cond_ric_alg.setter
    def qp_solver_cond_ric_alg(self, qp_solver_cond_ric_alg):
        if qp_solver_cond_ric_alg in [0, 1]:
            self.__qp_solver_cond_ric_alg = qp_solver_cond_ric_alg
        else:
            raise Exception(f'Invalid qp_solver_cond_ric_alg value. qp_solver_cond_ric_alg must be in [0, 1], got {qp_solver_cond_ric_alg}.')


    @qp_solver_cond_N.setter
    def qp_solver_cond_N(self, qp_solver_cond_N):
        if isinstance(qp_solver_cond_N, int) and qp_solver_cond_N >= 0:
            self.__qp_solver_cond_N = qp_solver_cond_N
        else:
            raise Exception('Invalid qp_solver_cond_N value. qp_solver_cond_N must be a positive int.')

    @qp_solver_cond_block_size.setter
    def qp_solver_cond_block_size(self, qp_solver_cond_block_size):
        if not isinstance(qp_solver_cond_block_size, list):
            raise Exception('Invalid qp_solver_cond_block_size value. qp_solver_cond_block_size must be a list of nonnegative integers.')
        for i in qp_solver_cond_block_size:
            if not isinstance(i, int) or not i >= 0:
                raise Exception('Invalid qp_solver_cond_block_size value. qp_solver_cond_block_size must be a list of nonnegative integers.')
        self.__qp_solver_cond_block_size = qp_solver_cond_block_size

    @qp_solver_warm_start.setter
    def qp_solver_warm_start(self, qp_solver_warm_start):
        if qp_solver_warm_start in [0, 1, 2]:
            self.__qp_solver_warm_start = qp_solver_warm_start
        else:
            raise Exception('Invalid qp_solver_warm_start value. qp_solver_warm_start must be 0 or 1 or 2.')

    @qp_tol.setter
    def qp_tol(self, qp_tol):
        if isinstance(qp_tol, float) and qp_tol > 0:
            self.__qp_solver_tol_eq = qp_tol
            self.__qp_solver_tol_ineq = qp_tol
            self.__qp_solver_tol_stat = qp_tol
            self.__qp_solver_tol_comp = qp_tol
        else:
            raise Exception('Invalid qp_tol value. qp_tol must be a positive float.')

    @qp_solver_tol_stat.setter
    def qp_solver_tol_stat(self, qp_solver_tol_stat):
        if isinstance(qp_solver_tol_stat, float) and qp_solver_tol_stat > 0:
            self.__qp_solver_tol_stat = qp_solver_tol_stat
        else:
            raise Exception('Invalid qp_solver_tol_stat value. qp_solver_tol_stat must be a positive float.')

    @qp_solver_tol_eq.setter
    def qp_solver_tol_eq(self, qp_solver_tol_eq):
        if isinstance(qp_solver_tol_eq, float) and qp_solver_tol_eq > 0:
            self.__qp_solver_tol_eq = qp_solver_tol_eq
        else:
            raise Exception('Invalid qp_solver_tol_eq value. qp_solver_tol_eq must be a positive float.')

    @qp_solver_tol_ineq.setter
    def qp_solver_tol_ineq(self, qp_solver_tol_ineq):
        if isinstance(qp_solver_tol_ineq, float) and qp_solver_tol_ineq > 0:
            self.__qp_solver_tol_ineq = qp_solver_tol_ineq
        else:
            raise Exception('Invalid qp_solver_tol_ineq value. qp_solver_tol_ineq must be a positive float.')

    @qp_solver_tol_comp.setter
    def qp_solver_tol_comp(self, qp_solver_tol_comp):
        if isinstance(qp_solver_tol_comp, float) and qp_solver_tol_comp > 0:
            self.__qp_solver_tol_comp = qp_solver_tol_comp
        else:
            raise Exception('Invalid qp_solver_tol_comp value. qp_solver_tol_comp must be a positive float.')

    @tol.setter
    def tol(self, tol):
        if isinstance(tol, float) and tol > 0:
            self.__nlp_solver_tol_eq = tol
            self.__nlp_solver_tol_ineq = tol
            self.__nlp_solver_tol_stat = tol
            self.__nlp_solver_tol_comp = tol
        else:
            raise Exception('Invalid tol value. tol must be a positive float.')

    @nlp_solver_tol_stat.setter
    def nlp_solver_tol_stat(self, nlp_solver_tol_stat):
        if isinstance(nlp_solver_tol_stat, float) and nlp_solver_tol_stat > 0:
            self.__nlp_solver_tol_stat = nlp_solver_tol_stat
        else:
            raise Exception('Invalid nlp_solver_tol_stat value. nlp_solver_tol_stat must be a positive float.')

    @nlp_solver_tol_eq.setter
    def nlp_solver_tol_eq(self, nlp_solver_tol_eq):
        if isinstance(nlp_solver_tol_eq, float) and nlp_solver_tol_eq > 0:
            self.__nlp_solver_tol_eq = nlp_solver_tol_eq
        else:
            raise Exception('Invalid nlp_solver_tol_eq value. nlp_solver_tol_eq must be a positive float.')

    @nlp_solver_tol_ineq.setter
    def nlp_solver_tol_ineq(self, nlp_solver_tol_ineq):
        if isinstance(nlp_solver_tol_ineq, float) and nlp_solver_tol_ineq > 0:
            self.__nlp_solver_tol_ineq = nlp_solver_tol_ineq
        else:
            raise Exception('Invalid nlp_solver_tol_ineq value. nlp_solver_tol_ineq must be a positive float.')

    @nlp_solver_tol_min_step_norm.setter
    def nlp_solver_tol_min_step_norm(self, nlp_solver_tol_min_step_norm):
        if isinstance(nlp_solver_tol_min_step_norm, float) and nlp_solver_tol_min_step_norm >= 0.0:
            self.__nlp_solver_tol_min_step_norm = nlp_solver_tol_min_step_norm
        else:
            raise Exception('Invalid nlp_solver_tol_min_step_norm value. nlp_solver_tol_min_step_norm must be a positive float.')

    @nlp_solver_ext_qp_res.setter
    def nlp_solver_ext_qp_res(self, nlp_solver_ext_qp_res):
        if nlp_solver_ext_qp_res in [0, 1]:
            self.__nlp_solver_ext_qp_res = nlp_solver_ext_qp_res
        else:
            raise Exception('Invalid nlp_solver_ext_qp_res value. nlp_solver_ext_qp_res must be in [0, 1].')

    @rti_log_residuals.setter
    def rti_log_residuals(self, rti_log_residuals):
        if rti_log_residuals in [0, 1]:
            self.__rti_log_residuals = rti_log_residuals
        else:
            raise Exception('Invalid rti_log_residuals value. rti_log_residuals must be in [0, 1].')

    @rti_log_only_available_residuals.setter
    def rti_log_only_available_residuals(self, rti_log_only_available_residuals):
        if rti_log_only_available_residuals in [0, 1]:
            self.__rti_log_only_available_residuals = rti_log_only_available_residuals
        else:
            raise Exception('Invalid rti_log_only_available_residuals value. rti_log_only_available_residuals must be in [0, 1].')

    @nlp_solver_tol_comp.setter
    def nlp_solver_tol_comp(self, nlp_solver_tol_comp):
        if isinstance(nlp_solver_tol_comp, float) and nlp_solver_tol_comp > 0:
            self.__nlp_solver_tol_comp = nlp_solver_tol_comp
        else:
            raise Exception('Invalid nlp_solver_tol_comp value. nlp_solver_tol_comp must be a positive float.')

    @nlp_solver_max_iter.setter
    def nlp_solver_max_iter(self, nlp_solver_max_iter):

        if isinstance(nlp_solver_max_iter, int) and nlp_solver_max_iter >= 0:
            self.__nlp_solver_max_iter = nlp_solver_max_iter
        else:
            raise Exception('Invalid nlp_solver_max_iter value. nlp_solver_max_iter must be a nonnegative int.')

    @print_level.setter
    def print_level(self, print_level):
        if isinstance(print_level, int) and print_level >= 0:
            self.__print_level = print_level
        else:
            raise Exception('Invalid print_level value. print_level takes one of the values >=0.')

    @model_external_shared_lib_dir.setter
    def model_external_shared_lib_dir(self, model_external_shared_lib_dir):
        if isinstance(model_external_shared_lib_dir, str) :
            self.__model_external_shared_lib_dir = model_external_shared_lib_dir
        else:
            raise Exception('Invalid model_external_shared_lib_dir value. Str expected.' \
            + '.\n\nYou have: ' + type(model_external_shared_lib_dir) + '.\n\n')

    @model_external_shared_lib_name.setter
    def model_external_shared_lib_name(self, model_external_shared_lib_name):
        if isinstance(model_external_shared_lib_name, str) :
            if model_external_shared_lib_name[-3:] == '.so' :
                raise Exception('Invalid model_external_shared_lib_name value. Remove the .so extension.' \
            + '.\n\nYou have: ' + type(model_external_shared_lib_name) + '.\n\n')
            else :
                self.__model_external_shared_lib_name = model_external_shared_lib_name
        else:
            raise Exception('Invalid model_external_shared_lib_name value. Str expected.' \
            + '.\n\nYou have: ' + type(model_external_shared_lib_name) + '.\n\n')

    @exact_hess_constr.setter
    def exact_hess_constr(self, exact_hess_constr):
        if exact_hess_constr in [0, 1]:
            self.__exact_hess_constr = exact_hess_constr
        else:
            raise Exception('Invalid exact_hess_constr value. exact_hess_constr takes one of the values 0, 1.')

    @exact_hess_cost.setter
    def exact_hess_cost(self, exact_hess_cost):
        if exact_hess_cost in [0, 1]:
            self.__exact_hess_cost = exact_hess_cost
        else:
            raise Exception('Invalid exact_hess_cost value. exact_hess_cost takes one of the values 0, 1.')

    @exact_hess_dyn.setter
    def exact_hess_dyn(self, exact_hess_dyn):
        if exact_hess_dyn in [0, 1]:
            self.__exact_hess_dyn = exact_hess_dyn
        else:
            raise Exception('Invalid exact_hess_dyn value. exact_hess_dyn takes one of the values 0, 1.')

    @fixed_hess.setter
    def fixed_hess(self, fixed_hess):
        if fixed_hess in [0, 1]:
            self.__fixed_hess = fixed_hess
        else:
            raise Exception('Invalid fixed_hess value. fixed_hess takes one of the values 0, 1.')

    @ext_cost_num_hess.setter
    def ext_cost_num_hess(self, ext_cost_num_hess):
        if ext_cost_num_hess in [0, 1]:
            self.__ext_cost_num_hess = ext_cost_num_hess
        else:
            raise Exception('Invalid ext_cost_num_hess value. ext_cost_num_hess takes one of the values 0, 1.')

    @num_threads_in_batch_solve.setter
    def num_threads_in_batch_solve(self, num_threads_in_batch_solve):
        if isinstance(num_threads_in_batch_solve, int) and num_threads_in_batch_solve > 0:
            self.__num_threads_in_batch_solve = num_threads_in_batch_solve
        else:
            raise Exception('Invalid num_threads_in_batch_solve value. num_threads_in_batch_solve must be a positive integer.')

    def set(self, attr, value):
        setattr(self, attr, value)



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#

from ...freetime import FreeTime
from ...casadi_helpers import prepare_build_dir
from ..method import ExternalMethod, linear_coeffs, check_Js
from ...solution import OcpSolution
from ...sampling_method import SamplingMethod

import numpy as np
from casadi import external, vec, CodeGenerator, SX, Sparsity, MX, vcat, veccat, hcat, symvar, substitute, densify, sparsify, DM, Opti, is_linear, vertcat, depends_on, jacobian, linear_coeff, quadratic_coeff, mtimes, pinv, evalf, Function, vvcat, inf, sum1, sum2, diag, solve, fmin, fmax
import casadi
import casadi as cs
from ...casadi_helpers import DM2numpy, reshape_number
from collections import OrderedDict

import subprocess
import os
from ctypes import *
import glob
import shutil
import platform
import hashlib
from casadi import Callback

windows_generator = None
if platform.system()=="Windows":
    if "VCVARS" in os.environ:
        vcvars = os.environ["VCVARS"]
    else:
        vcvars = None
        def vcvars_dirs():
            for p in ["Program Files (x86)","Program Files"]:
                for suffix in ["","Community\\","Professional\\"]:
                    yield f"C:\\{p}\\Microsoft Visual Studio\\2022\\{suffix}VC\Auxiliary\\Build\\vcvars64.bat"
                    yield f"C:\\{p}\\Microsoft Visual Studio\\2019\\{suffix}VC\Auxiliary\\Build\\vcvars64.bat"
                    yield f"C:\\{p}\\Microsoft Visual Studio\\2017\\{suffix}VC\Auxiliary\\Build\\vcvars64.bat"
        for e in vcvars_dirs():
            if os.path.exists(e):
                vcvars = e
                break
        if vcvars is None:
            raise Exception("Could not find vcvars bat file. Tried %s." % str(list(vcvars_dirs())))

        if "2022" in vcvars:
            windows_generator = "Visual Studio 17 2022"
        if "2019" in vcvars:
            windows_generator = "Visual Studio 16 2019"
        if "2017" in vcvars:
            windows_generator = "Visual Studio 15 2017"

def get_terms(e):
    def get_terms_internal(e):
        if e.op()==cs.OP_ADD:
            for i in range(e.n_dep()):
                for t in get_terms_internal(e.dep(i)):
                    yield t
        else:
            yield e
    return list(get_terms_internal(e))


def visit(e,parents=None):
    if parents is None:
        parents = []
    yield (e,parents)
    for i in range(e.n_dep()):
        for t in visit(e.dep(i),parents=[e]+parents):
            yield t

"""
conda install -c conda-forge make cmake
"""

external_dir = os.path.join(os.path.dirname(__file__),"external")

#if not os.path.exists(external_dir):
#    raise Exception("FATROP source not found")

# def run_build(cwd=None):
#     subprocess.run(["bash","build.sh"],cwd=cwd)

INF = inf



stats_fields = [('stop_crit',c_int),('conv_grad',c_int),('conv_con',c_int),('n_outer_iter',c_int),('n_inner_iter',c_int),('runtime',c_double)]

class StatsStruct(Structure):
    _fields_ = stats_fields


def format_float(e):
    return "%0.18f" % e

def strlist(a):
    elems = []
    for e in a:
        if isinstance(e,str):
            elems.append('"'+e+'"')
        elif isinstance(e,float):
            elems.append(format_float(e))
        else:
            elems.append(str(e))
    return ",".join(elems)
    
def check_Js(J):
    """
    Checks if J, a pre-multiplier for slacks, is of legitimate structure
    Empty rows are allowed
    """
    try:
        J = evalf(J)
    except:
        raise Exception("Slack error")
    assert np.all(np.array(J.nonzeros())==1), "All nonzeros must be 1"
    # Check if slice of permutation of unit matrix
    assert np.all(np.array(sum2(J))<=1), "Each constraint can only depend on one slack at most"
    assert np.all(np.array(sum1(J))<=1), "Each constraint must depend on a unique slack, if any"

def fill_in(var, expr,value, array, k):
    value = evalf(value)
    J, r = linear_coeffs(var,expr)
    J = evalf(J)
    r = evalf(r)
    assert r.is_zero()
    check_Js(J)
    expr = reshape_number(var, expr)
    if J.sparsity().get_col():
        array[J.sparsity().get_col(), k] = value[J.row()]
def fill_in_array(var, expr, value, array, numeric = True):
    if numeric:
        value = evalf(value)
    J, r = linear_coeffs(var[:,0],expr)
    J = evalf(J)
    r = evalf(r)
    assert r.is_zero()
    check_Js(J)
    expr = reshape_number(var, expr)
    if J.sparsity().get_col():
        array[J.sparsity().get_col(), :] = value[J.row(), :]
def get_offsets(var, expr):
    J, r = linear_coeffs(var,expr)
    J = evalf(J)
    r = evalf(r)
    assert r.is_zero()
    check_Js(J)
    expr = reshape_number(var, expr)
    return (J.row(), J.sparsity().get_col())

def mark(slack, Js):
    assert np.all(np.array(slack[Js.sparsity().get_col()])==0)
    slack[Js.sparsity().get_col()] = 1

def export_expr(m):
    if isinstance(m,list):
        if len(m)==0:
            return MX(0, 1)
        else:
            return vcat(m)
    return m

def export_num(m):
    res=np.array(evalf(export_expr(m)))
    if np.any(res==-inf) or np.any(res==inf):
        print("WARNING: Double-sided constraints are much preferred. Replaced inf with %f." % INF)
    res[res==-inf] = -INF
    res[res==inf] = INF
    return res

def export_num_vec(m):
    return np.array(evalf(export_expr(m))).reshape(-1)

def export(m):
    return (export_expr(m),False)

def export_vec(m):
    return (export_expr(m),True)


class SourceArtifact:
    def __init__(self, name):
        self.name = name

class LibraryArtifact:
    def __init__(self, basename, relpath):
        self.basename = basename # No starting lib, ending so/dll
        self.relpath = relpath

class FatropMethod(ExternalMethod):
    def __init__(self,
        verbose=False,
        fatrop_options=None,
        intg= 'rk',
        M = 1,
        mode = "fatropy",
        **kwargs):
        self.build_dir_abs = "./foobar"
        supported = {}
        ExternalMethod.__init__(self, supported=supported, **kwargs)
        self.fatrop_options = {} if fatrop_options is None else fatrop_options
        our_defaults = {}
        for k,v in our_defaults.items():
            if k not in self.fatrop_options:
                self.fatrop_options[k] = v
        self.codegen_name = 'casadi_codegen'
        self.user = "((cs_struct*) userparam)"
        self.user_fatrop = "((cs_struct*) fatrop->userparam)"
        self.Nhor = self.N+1
        self.intg = intg
        self.M = M #Assumption!
        # self.N = N
        self.verbose = verbose
        self.artifacts = []
        self.poly_coeff = []  # Optional list to save the coefficients for a polynomial
        self.poly_coeff_z = []  # Optional list to save the coefficients for a polynomial
        self.xk = []  # List for intermediate integrator states
        self.zk = []
        self.samplers = {}
        self.variable_names = {}
        self.mode = mode 
        
    def set_name(self, name):
        self.build_dir_abs = "./"+name 
    def set_expand(self, expand):
        self.expand = expand
    def regname(self, symbol_in, name):
        self.variable_names[symbol_in] = name

    def fill_placeholders_integral(self, phase, stage, expr, *args):
        raise Exception("ocp.integral not supported. Use ocp.sum instead.")
        # if phase==1:
        #     return expr

    def fill_placeholders_sum_control(self, phase, stage, expr, *args):
        if phase == 1:
            return expr
        # raise Exception("ocp.sum not supported. Use ocp.integral instead.")

    def fill_placeholders_sum_control_plus(self, phase, stage, expr, *args):
        if phase == 1:
            return expr

    def fill_placeholders_DT_discrete(self, phase, stage, expr, *args):
        if phase == 1:
            if not stage._state_next:
                raise Exception("Discrete time integrator DT found but dynamics ode is continuous time") 
        return None

    def _register(self,fun_name,argtypes,restype):
        self.prefix=""
        fun = getattr(self.lib,self.prefix+fun_name)
        setattr(self,"_"+fun_name,fun)
        fun.argtypes = argtypes
        fun.restype = restype

    def transcribe_phase1(self, stage, **kwargs):
        # It is not really transcription because FATROP simply takes stage-wise costs as input. Not a large NLP as input.
        #Phase 1 deals with creating placeholder variables and objectives

        self.preamble = ["casadi_int sz_arg=0, sz_res=0, sz_iw=0, sz_w=0;",
                         "casadi_int sz_arg_local, sz_res_local, sz_iw_local, sz_w_local;",
                        ]
        self.postamble = []

        self.stage = stage
        self.opti = Opti()

        # Is computing the whole grid needed for FATROP?
        ## self.time_grid = self.grid(stage._t0, stage._T, self.N)
        self.normalized_time_grid = self.grid(0.0, 1.0, self.N)
        self.time_grid = self.normalized_time_grid
        if self.t_state:
            if isinstance(stage._T, FreeTime):
                stage.set_initial(stage.t, self.time_grid*stage._T.T_init)
            else:
                stage.set_initial(stage.t, self.time_grid*stage._T)
        if not isinstance(stage._T, FreeTime): self.time_grid*= stage._T
        if not isinstance(stage._t0, FreeTime): self.time_grid+= stage._t0
        self.control_grid = MX(stage.t0 + self.normalized_time_grid*stage.T)

        if not isinstance(stage._T, FreeTime) and not isinstance(stage._t0, FreeTime):
            dT = stage._T/(self.N)
        else:
            dT = self.T/(self.N)
            stage.subject_to(stage.at_t0(self.T)>0)
            stage.set_initial(self.T, stage._T.T_init)

        # f = stage._ode()
        f = SamplingMethod(intg = self.intg, M = self.M).discrete_system(stage)
        if self.expand:
            f = f.expand()
        self.f = f

        xkp1 = MX.sym('xkp1', stage.nx)
        x_sym = stage.x # MX.sym('x', stage.nx)
        u_sym = stage.u #MX.sym('u', stage.nu)
        global_params_sym = stage.p #MX.sym('p', stage.np)
        stage_params_sym = veccat(*(stage.parameters['control'] + stage.parameters['control+']))
        self.stage_params_sym = stage_params_sym
        global_params_sym = veccat(*stage.parameters[''])
        self.global_params_sym = global_params_sym
        x_next = f(x0=x_sym, u=u_sym, T=dT, p=stage.p)["xf"]
        self.x_next = x_next
        b_x = - xkp1 + x_next

        options = {}
        options["with_header"] = True
        self.codegen = CodeGenerator(f"{self.codegen_name}.c", options)

        BAbt = cs.jacobian(x_next, veccat(u_sym, x_sym)).T
        BAbt = cs.vertcat(BAbt, b_x.T)

        self.artifacts.append(SourceArtifact(f"{self.codegen_name}.c"))
        self.artifacts.append(SourceArtifact(f"{self.codegen_name}.h"))
        self.artifacts.append(LibraryArtifact("fatrop_driver", "lib"))


        # self.artifacts.append(SourceArtifact(f"{self.fatrop_driver}.cpp"))
        # self.wrapper = Wrapper(userparam=self.user_fatrop)

        assert len(stage.variables['control'])==0, "variables defined on control grid not supported. Use controls instead."
        self.v = cs.MX.zeros(0,1)

        self.p_global = cs.vvcat(stage.parameters[''])
        self.p_stage = stage.parameters['control'] + stage.parameters['control+']
        self.X_gist = [cs.vvcat([MX.sym("Xg", s.size1(), s.size2()) for s in stage.states]) for k in range(self.N+1)]
        self.U_gist = [cs.vvcat([MX.sym("Ug", s.size1(), s.size2()) for s in stage.controls]) for k in range(self.N)]
        self.P_stage_gist = [cs.vvcat([MX.sym("stage_param", s.size1(), s.size2()) for s in  self.p_stage]) for k in range(self.N+1)]
        self.P_global_gist = cs.vvcat([MX.sym("Vg", s.size1(), s.size2()) for s in stage.parameters['']])
        self.T_gist = MX.sym("Tg")

        self.x = cs.vvcat(self.X_gist+self.U_gist)

        self.gist_parts = []
        self.gist_parts.append((self.X_gist, stage.x, "local"))
        self.gist_parts.append((self.U_gist, stage.u, "local"))
        self.gist_parts.append((self.P_global_gist, self.p_global_cat, "global"))
        self.gist_parts.append((self.P_stage_gist, self.p_local_cat, "local"))
        self.gist_parts.append((self.T_gist, stage.T, "global"))

        self.X = self.opti.variable(*stage.x.shape)
        self.U = self.opti.variable(*stage.u.shape)
        self.V = self.opti.variable(*self.v.shape)
        self.P = self.opti.parameter(*stage.p.shape)
        self.t = self.opti.parameter()
        self.T = self.opti.variable()

        self.raw = [stage.x,stage.u,stage.p,stage.t, self.v]
        self.optivar = [self.X, self.U, self.P, self.t, self.V]
        if self.free_time:
            self.raw += [stage.T]
            self.optivar += [self.T]

        

        inits = []
        inits.append((stage.T, stage._T.T_init if isinstance(stage._T, FreeTime) else stage._T))
        inits.append((stage.t0, stage._t0.T_init if isinstance(stage._t0, FreeTime) else stage._t0))

        self.inits = inits

        self.control_grid_init = evalf(substitute([self.control_grid], [a for a,b in inits],[b for a,b in inits])[0])

        #self.control_grid = self.normalized_time_grid

        self.lagrange = MX(0)
        self.mayer = MX(0)
        self.objI = MX(0)
        var_mayer = []
        obj = MX(stage._objective)
        terms = get_terms(obj)
        for term in terms:
            n = [e.name() for e in symvar(term)]
            sumi = np.sum([e=="r_sum_control" for e in n])
            sumip = np.sum([e=="r_sum_control_plus" for e in n])
            summ = np.sum([e=="r_at_tf" for e in n])
            summI = np.sum([e=="r_at_t0" for e in n])
            if sumi + sumip + summ  + summI!= 1:
                raise Exception("Objective cannot be parsed")
            if sumi==1 or sumip ==1:
                n_hits = 0
                for e,parents in visit(term):
                    if e.is_symbolic() and e.name()=="r_sum_control":
                        n_hits+=1
                        for pi in range(len(parents)):
                            p = parents[pi]
                            pchild = parents[pi-1] if pi>0 else e
                            correct = False
                            # Distributive operations
                            if p.op()==cs.OP_MUL:
                                correct = True
                            if p.op()==cs.OP_DIV:
                                # Only allow integral in LHS
                                correct = hash(p.dep(0))==hash(pchild)
                            assert correct, "Objective cannot be parsed: integrals can only be multiplied or divided."
                    if n_hits>1:
                        raise Exception("Objective cannot be parsed")
                if summ!=0:
                    raise Exception("Objective cannot be parsed: operation combining integral and at_tf part not supported")
                self.lagrange += term
                self.objI += term
            if sumip==1 or summ ==1:
                self.mayer += term
            if summI ==1:
                self.objI += term 
        self.P0 = DM.zeros(stage.np)

    def discrete_system(self, stage):
        # Added stage as argument to make it compatible with Stage.discrete_system()
        return self.f

    def set_value(self, stage, master, parameter, value):
        value = evalf(value)
        if parameter in stage.parameters['']:
            p = vec(parameter)
            fill_in(p, self.global_params_sym, value, self.global_params_value,0)
        else:
            p = vec(parameter)
            # for k in list(range(self.N)):
            if p.numel()==value.numel():
                if(value.shape[1] == p.shape[0]):
                    value = value.T
                value = cs.repmat(value,  1, self.N+1)
            if p.numel()*(self.N)==value.numel() or p.numel()*(self.N+1)== value.numel():
                if(value.shape[1] == p.shape[0]):
                    value = value.T
            assert(value.shape[0] == p.shape[0])
            if p.numel()*(self.N)== value.numel():
                value = cs.horzcat(value, value[:,-1])
            fill_in_array(p, self.stage_params_sym, value, self.stage_params_value)


        ## TODO this seems to work but is this the right way to do it?
        # master.stage._param_vals[parameter] = value
    def get_parameters(self, stage):
        self.global_params_value = DM.zeros(self.global_params_sym.shape)
        self.stage_params_value = DM.zeros(self.stage_params_sym.shape[0], self.N+1)
        for i, p in enumerate(stage.parameters['']):
            self.set_value(stage, None, p, stage._param_value(p))
        for i, p in enumerate(stage.parameters['control'] + stage.parameters['control+']):
            self.set_value(stage,None,  p, stage._param_value(p))
        return self.global_params_value, self.stage_params_value
    def set_initial(self, arg1, arg2, initial_dict):
        self.U0, self.X0 = self.get_initial(self.stage, initial_dict)
        if self.mode == "fatropy":
            self.myOCP.set_initial_u(self.U0)
            self.myOCP.set_initial_x(self.X0)
        pass
    def add_sampler(self, name, expr):
        self.samplers[name] = expr

    def _register(self,fun_name,argtypes,restype):
        self.prefix=""
        fun = getattr(self.lib,self.prefix+fun_name)
        setattr(self,"_"+fun_name,fun)
        fun.argtypes = argtypes
        fun.restype = restype

    def get_initial(self, stage, initial_dict):

        U0= DM.zeros(stage.nu, self.N)
        X0 = DM.zeros(stage.nx, self.N+1)
        for var, expr in initial_dict.items():
                # print(var, expr)
                var = vec(var)
                expr = expr
                value = evalf(expr)
                # value = vec(value)
                if depends_on(var,stage.u):
                    assert not depends_on(var, self.v)
                    if var.numel()==value.numel():
                        value = vec(value)
                        if(value.shape[1] == var.shape[0]):
                            value = value.T
                        value = cs.repmat(value,  1, self.N)
                    if var.numel()*(self.N)==value.numel():
                        if(value.shape[1] == var.shape[0]):
                            value = value.T
                    if var.numel()*(self.N+1)==value.numel():
                        if(value.shape[1] == var.shape[0]):
                            value = value.T
                        value = value[:,:-1]
                    assert(value.shape[0] == var.shape[0])
                    fill_in_array(var, stage.u, value, U0)
                if depends_on(var,stage.x):
                    assert not depends_on(vec(var), self.v)
                    if var.numel()==value.numel():
                        value = vec(value)
                        if(value.shape[1] == var.shape[0]):
                            value = value.T
                        value = cs.repmat(value,  1, self.N+1)
                    if var.numel()*(self.N+1)==value.numel():
                        if(value.shape[1] == var.shape[0]):
                            value = value.T
                    assert(value.shape[0] == var.shape[0])
                    fill_in_array(var, stage.x, value, X0)
        return U0, X0

    # def eval_expr(expr, )
    def transcribe_phase2(self, stage, build_dir_abs=None,**kwargs):

        if build_dir_abs is None: build_dir_abs = self.build_dir_abs
        print("build_dir_abs", build_dir_abs, kwargs)
        # Phase 2 adds the constraints
        
        opti_advanced = self.opti.advanced
        placeholders = kwargs["placeholders"]



        # Total Lagrange integrand
        lagrange = placeholders(self.lagrange,preference=['expose'])
        # Total Mayer term
        mayer = placeholders(self.mayer,preference=['expose'])
        objI = placeholders(self.objI,preference=['expose'])
        self.x_next = placeholders(self.x_next, preference='expose')

        x0_eq = []
        x0_b = []

        eq_init = [MX.zeros(0)]
        eq_mid = [MX.zeros(0)] #
        eq_term = [MX.zeros(0)]
        ineq_init = [MX.zeros(0)]
        ineq_mid = [MX.zeros(0)] # <=0
        ineq_term = [MX.zeros(0)]
        ub_init = [MX.zeros(0)]
        lb_init = [MX.zeros(0)]
        ub_mid = [MX.zeros(0)]
        lb_mid = [MX.zeros(0)]
        ub_term = [MX.zeros(0)]
        lb_term = [MX.zeros(0)]  

        # Process initial point constraints
        # Probably should de-duplicate stuff wrt path constraints code
        for c, meta, _ in stage._constraints["point"]:
            # Make sure you resolve u to r_at_t0/r_at_tf

            if not 'r_at_t0' in [a.name() for a in symvar(c)]:
                continue
            c = placeholders(c,max_phase=1)

            cb = c
            c = substitute([placeholders(c,preference='expose')],self.raw,self.optivar)[0]
            mc = opti_advanced.canon_expr(c) # canon_expr should have a static counterpart
            lb,canon,ub = substitute([mc.lb,mc.canon,mc.ub],self.optivar,self.raw)
            check = not depends_on(canon, veccat(self.v))
            assert check, 'v variables are not supported yet'
            if mc.type == casadi.OPTI_EQUALITY:
                eq_init.append(canon-ub)
            else:
                assert mc.type in [casadi.OPTI_INEQUALITY, casadi.OPTI_GENERIC_INEQUALITY, casadi.OPTI_DOUBLE_INEQUALITY]
                ineq_init.append(canon)
                ub_init.append(ub)
                lb_init.append(lb)
        
        # Process terminal point constraints
        # Probably should de-duplicate stuff wrt path constraints code
        for c, meta, _ in stage._constraints["point"]:
            # Make sure you resolve u to r_at_t0/r_at_tf

            if not 'r_at_tf' in [a.name() for a in symvar(c)]:
                continue
            c = substitute([placeholders(c,preference='expose')],self.raw,self.optivar)[0]
            mc = opti_advanced.canon_expr(c) # canon_expr should have a static counterpart
            lb,canon,ub = substitute([mc.lb,mc.canon,mc.ub],self.optivar,self.raw)

            check = not depends_on(canon, veccat(stage.u, self.v))
            assert check, "at t=tF, only constraints on x are allowed. Got '%s'" % str(c)

            if mc.type == casadi.OPTI_EQUALITY or mc.type == casadi.OPTI_GENERIC_EQUALITY:
                eq_term.append(canon-ub)
            else:
                assert mc.type in [casadi.OPTI_INEQUALITY, casadi.OPTI_GENERIC_INEQUALITY, casadi.OPTI_DOUBLE_INEQUALITY]
                ineq_term.append(canon)
                ub_term.append(ub)
                lb_term.append(lb)
        

        # Process path constraints
        # TODO check for include first and include last
        for c, meta, args in stage._constraints["control"]+stage._constraints["integrator"]:
            c = substitute([placeholders(c,preference=['expose'])],self.raw,self.optivar)[0]
            mc = opti_advanced.canon_expr(c) # canon_expr should have a static counterpart
            lb,canon,ub = substitute([mc.lb,mc.canon,mc.ub],self.optivar,self.raw)
            check = not depends_on(canon, veccat(self.v))
            assert check, 'v variables'
            # lb <= canon <= ub
            # Check for infinities
            try:
                lb_inf = np.all(np.array(evalf(lb)==-inf))
            except:
                lb_inf = False
            try:
                ub_inf = np.all(np.array(evalf(ub)==inf))
            except:
                ub_inf = False

            if mc.type == casadi.OPTI_EQUALITY or mc.type == casadi.OPTI_GENERIC_EQUALITY:
                if mc.type == casadi.OPTI_GENERIC_EQUALITY:
                    canon = lb-ub
                    lb = DM(canon.sparsity())
                    ub = DM(canon.sparsity())
                eq_mid.append(canon-ub)
                if args['include_first']:
                    eq_init.append(canon-ub)
                if args['include_last']:
                    check = not depends_on(canon, veccat(stage.u, self.v))
                    if not check and not depends_on(canon, veccat(stage.x)): continue
                    assert check, "at t=tF, only constraints on x are allowed. Got '%s'" % str(c)
                    eq_term.append(canon-ub)
            else:
                assert mc.type in [casadi.OPTI_INEQUALITY, casadi.OPTI_GENERIC_INEQUALITY, casadi.OPTI_DOUBLE_INEQUALITY]

                ineq_mid.append(canon)
                ub_mid.append(ub)
                lb_mid.append(lb)
                if args['include_first']:
                    ineq_init.append(canon)
                    ub_init.append(ub)
                    lb_init.append(lb)
                if args['include_last']:
                    check = not depends_on(canon, veccat(stage.u, self.v))
                    if not check and not depends_on(canon, veccat(stage.x)): continue
                    assert check, "at t=tF, only constraints on x are allowed. Got '%s'" % str(c)
                    ineq_term.append(canon)
                    ub_term.append(ub)
                    lb_term.append(lb)
        


        #### parameters
        global_params_value, stage_params_value =self.global_params_value, self.stage_params_value = self.get_parameters(stage)

        #### initializiaton
        U0, X0 = self.U0, self.X0 = self.get_initial(stage, stage._initial)


        eqI = vvcat(eq_init)
        eq = vvcat(eq_mid)
        eqF = vvcat(eq_term)
        ineqI = vvcat(ineq_init)
        ineq = vvcat(ineq_mid)
        ineqF = vvcat(ineq_term)
        lb_init = vvcat(lb_init)
        ub_init = vvcat(ub_init)
        lb_mid = vvcat(lb_mid)
        ub_mid = vvcat(ub_mid)
        ub_term = vvcat(ub_term)
        lb_term = vvcat(lb_term)

        ngI = eqI.shape[0]
        ng = eq.shape[0]
        ngF = eqF.shape[0]
        ngIneqI = ineqI.shape[0]
        ngIneq = ineq.shape[0]
        ngIneqF = ineqF.shape[0]
        dual_dyn = MX.sym('d_dyn', stage.nx)
        dual_eqI = MX.sym('d_eqI', ngI)
        dual_eq = MX.sym('d_eq', ng)
        dual_eqF = MX.sym('d_eqF', ngF)
        dualIneqI = MX.sym('d_IneqI', ngIneqI)
        dualIneq = MX.sym('d_Ineq', ngIneq)
        dualIneqF = MX.sym('d_IneqF', ngIneqF)
        nx = stage.nx
        nu = stage.nu
        def expand_help(func):
            if self.expand:
                return func.expand()
            else:
                return func


        #TODO: Stagewise equality constraints seem to be missing in FATROP?

        obj_scale = MX.sym('obj_scale')
                # BAbt
        stateskp1 = MX.sym("states_kp1", nx)
        BAbt = MX.zeros(nu+nx+1, nx)
        BAbt[:nu+nx,
             :] = jacobian(self.x_next, veccat(stage.u, stage.x)).T
        b = (-stateskp1 + self.x_next)[:]
        BAbt[nu+nx, :] = b
        self.codegen.add(
            expand_help(Function("BAbt", [stateskp1, stage.u, stage.x, self.stage_params_sym, self.global_params_sym], [densify(BAbt)])))
        # b
        self.codegen.add(expand_help(Function("bk", [stateskp1, stage.u,
                              stage.x, self.stage_params_sym, self.global_params_sym], [densify(b)])))
        # RSQrqtI
        RSQrqtI = MX.zeros(nu+nx+1, nu + nx)
        [RSQI, rqI] = cs.hessian(objI, veccat(stage.u, stage.x))
        RSQIGN = RSQI
        rqlagI = rqI
        if ngI > 0:
            [H, h]= cs.hessian(dual_eqI.T@eqI, veccat(stage.u, stage.x))
            RSQI += H
            rqlagI += h
        [H,h] = cs.hessian(dual_dyn.T@self.x_next,
                        veccat(stage.u, stage.x))
        RSQI += H
        rqlagI += h
        
        if ngIneqI > 0:
            [H,h] = cs.hessian(dualIneqI.T@ineqI,
                            veccat(stage.u, stage.x))
            RSQI += H
            rqlagI += h
        RSQrqtI[:nu+nx, :] = RSQI
        RSQrqtI[nu+nx, :] = rqlagI[:]
        self.codegen.add(expand_help(Function("RSQrqtI", [obj_scale, stage.u,
              stage.x, dual_dyn, dual_eqI, dualIneqI, self.stage_params_sym, self.global_params_sym], [densify(RSQrqtI)])))
        RSQrqtI[:nu+nx, :] = RSQIGN
        RSQrqtI[nu+nx, :] = rqlagI[:]
        self.codegen.add(expand_help(Function("RSQrqtIGN", [obj_scale, stage.u,
              stage.x, dual_dyn, dual_eqI, dualIneqI, self.stage_params_sym, self.global_params_sym], [densify(RSQrqtI)])))
        # rqI
        self.codegen.add(expand_help(Function("rqI", [obj_scale,
              stage.u, stage.x, self.stage_params_sym, self.global_params_sym], [densify(rqI)])))
        # RSQrqt
        RSQrqt = MX.zeros(nu+nx+1, nu + nx)
        [RSQ, rq] = cs.hessian(lagrange, veccat(stage.u, stage.x))
        RSQGN = RSQ
        rqlag = rq
        if ng > 0:
            [H, h]= cs.hessian(dual_eq.T@eq, veccat(stage.u, stage.x))
            RSQ += H
            rqlag += h
        [H,h]= cs.hessian(dual_dyn.T@self.x_next,
                       veccat(stage.u, stage.x))
        RSQ += H
        rqlag +=h

        if ngIneq > 0:
            [H,h] = cs.hessian(dualIneq.T@ineq,
                           veccat(stage.u, stage.x))
            RSQ += H
            rqlag +=h
        RSQrqt[:nu+nx, :] = RSQ
        RSQrqt[nu+nx, :] = rqlag[:]
        self.codegen.add(expand_help(Function("RSQrqt", [obj_scale, stage.u, stage.x,
              dual_dyn, dual_eq, dualIneq, self.stage_params_sym, self.global_params_sym], [densify(RSQrqt)])))
        RSQrqt[:nu+nx, :] = RSQGN
        RSQrqt[nu+nx, :] = rqlag[:]
        self.codegen.add(expand_help(Function("RSQrqtGN", [obj_scale, stage.u, stage.x,
              dual_dyn, dual_eq, dualIneq, self.stage_params_sym, self.global_params_sym], [densify(RSQrqt)])))
        # rqF
        self.codegen.add(expand_help(Function("rqk", [obj_scale,
              stage.u, stage.x, self.stage_params_sym, self.global_params_sym], [densify(rq)])))
        # Lk
        self.codegen.add(expand_help(Function("LI", [obj_scale, stage.u,
              stage.x, self.stage_params_sym, self.global_params_sym], [densify(objI)])))
        # Lk
        self.codegen.add(expand_help(Function("Lk", [obj_scale, stage.u,
              stage.x, self.stage_params_sym, self.global_params_sym], [densify(lagrange)])))
        # RSQrqtF
        RSQrqtF = MX.zeros(nx+1, nx)
        [RSQF, rqF] = cs.hessian(mayer, veccat(stage.x))
        RSQFGN = RSQF
        rqlagF = rqF
        if ngF > 0:
            [H, h]= cs.hessian(dual_eqF.T@eqF,
                            veccat(stage.x))
            RSQF += H
            rqlagF += h
        if ngIneqF > 0:
            [H,h] = cs.hessian(dualIneqF.T@ineqF,
                           veccat(stage.x))
            RSQF += H
            rqlagF += h
        # if ngIneq>-1:
        #     RSQF += cs.hessian(dualIneq.T@ineq, vertcat(stage.u, stage.x))[-1]
        RSQrqtF[:nx, :] = RSQF
        RSQrqtF[nx, :] = rqlagF[:]
        self.codegen.add(expand_help(Function("RSQrqtF", [obj_scale, stage.u, stage.x,
              dual_dyn, dual_eqF, dualIneqF, self.stage_params_sym, self.global_params_sym], [densify(RSQrqtF)])))
        RSQrqtF[:nx, :] = RSQFGN
        RSQrqtF[nx, :] = rqlagF[:]
        self.codegen.add(expand_help(Function("RSQrqtFGN", [obj_scale, stage.u, stage.x,
              dual_dyn, dual_eqF, dualIneqF, self.stage_params_sym, self.global_params_sym], [densify(RSQrqtF)])))
        # rqF
        self.codegen.add(expand_help(Function("rqF", [obj_scale,
              stage.u, stage.x, self.stage_params_sym, self.global_params_sym], [densify(rqF)])))
        # LF
        self.codegen.add(expand_help(Function("LF", [obj_scale, stage.u,
              stage.x, self.stage_params_sym, self.global_params_sym], [densify(mayer)])))
        # GgtI
        GgtI = MX.zeros(nu+nx+1, ngI)
        GgtI[:nu+nx,
             :] = jacobian(eqI, veccat(stage.u, stage.x)).T
        GgtI[nu+nx, :] = eqI[:].T
        self.codegen.add(expand_help(Function(
            "GgtI", [stage.u, stage.x, self.stage_params_sym, self.global_params_sym], [densify(GgtI)])))
        # g_I
        self.codegen.add(expand_help(Function("gI", [stage.u, stage.x, self.stage_params_sym,
              self.global_params_sym], [densify(eqI[:])])))
        # Ggt
        Ggt = MX.zeros(nu+nx+1, ng)
        Ggt[:nu+nx,
             :] = jacobian(eq, veccat(stage.u, stage.x)).T
        Ggt[nu+nx, :] = eq[:].T
        self.codegen.add(expand_help(Function(
            "Ggt", [stage.u, stage.x, self.stage_params_sym, self.global_params_sym], [densify(Ggt)])))
        # g
        self.codegen.add(expand_help(Function("g", [stage.u, stage.x, self.stage_params_sym,
              self.global_params_sym], [densify(eq[:])])))
        # GgtF
        GgtF = MX.zeros(nx+1, ngF)
        GgtF[:nx, :] = jacobian(eqF, stage.x).T
        GgtF[nx, :] = eqF[:].T
        self.codegen.add(expand_help(Function(
            "GgtF", [stage.u, stage.x, self.stage_params_sym, self.global_params_sym], [densify(GgtF)])))
        # g_F
        self.codegen.add(expand_help(Function("gF", [stage.u, stage.x, self.stage_params_sym,
              self.global_params_sym], [densify(eqF[:])])))
        # GgineqIt
        GgineqIt = MX.zeros(nu+nx+1, ngIneqI)
        GgineqIt[:nu+nx,
                :] = jacobian(ineqI, veccat(stage.u, stage.x)).T
        GgineqIt[nu+nx, :] = ineqI[:].T
        self.codegen.add(expand_help(Function("GgineqIt", [stage.u,
              stage.x, self.stage_params_sym, self.global_params_sym], [densify(GgineqIt)])))
        self.codegen.add(expand_help(Function("gineqI", [stage.u, stage.x, self.stage_params_sym, self.global_params_sym], [
              densify(ineqI[:])])))
        # Ggineqt
        Ggineqt = MX.zeros(nu+nx+1, ngIneq)
        Ggineqt[:nu+nx,
                :] = jacobian(ineq, veccat(stage.u, stage.x)).T
        Ggineqt[nu+nx, :] = ineq[:].T
        self.codegen.add(expand_help(Function("Ggineqt", [stage.u,
              stage.x, self.stage_params_sym, self.global_params_sym], [densify(Ggineqt)])))
        self.codegen.add(expand_help(Function("gineq", [stage.u, stage.x, self.stage_params_sym, self.global_params_sym], [
              densify(ineq[:])])))
        # GgineqFt
        GgineqFt = MX.zeros(nx+1, ngIneqF)
        GgineqFt[:nx,
                :] = jacobian(ineqF, veccat(stage.x)).T
        GgineqFt[nx, :] = ineqF[:].T
        self.codegen.add(expand_help(Function("GgineqFt", [
              stage.x, self.stage_params_sym, self.global_params_sym], [densify(GgineqFt)])))
        self.codegen.add(expand_help(Function("gineqF", [stage.x, self.stage_params_sym, self.global_params_sym], [
              densify(ineqF[:])])))
        
        # codegenerate samplers
        for sampl in self.samplers.keys():
            self.codegen.add(expand_help(Function("sampler_"+ str(sampl), [stage.u, stage.x, self.stage_params_sym, self.global_params_sym], [densify(placeholders(self.samplers[sampl], preference =['expose']))])))


        self.build_dir_abs = build_dir_abs
        # prepare_build_dir(self.build_dir_abs)
        # check if the build directory exists otherwise create it
        if not os.path.exists(self.build_dir_abs):
            os.makedirs(self.build_dir_abs)

        self.codegen.generate(self.build_dir_abs+os.sep)
        def get_sym_name(symbol):
            # return symbol.name() if bool(symbol not in self.variable_names) else self.variable_names[symbol]
            return symbol.name()

        # creating the json file for fatrop
        control_params_sym = stage.parameters['control'] + stage.parameters['control+']
        global_params_sym = stage.parameters['']
        control_params_offset = {get_sym_name(sym):get_offsets(vec(sym), veccat(*control_params_sym)) for sym in control_params_sym}
        global_params_offset = {get_sym_name(sym):get_offsets(vec(sym), veccat(*global_params_sym)) for sym in global_params_sym}
        states_sym = [vec(state) for state in  stage.states]
        states_offset = {get_sym_name(sym):get_offsets(vec(sym), veccat(*states_sym)) for sym in stage.states}
        controls_sym = [vec(control) for control in  stage.controls]
        controls_offset = {get_sym_name(sym):get_offsets(vec(sym), veccat(*controls_sym)) for sym in stage.controls}



        json_dict = {
                        'control_params_offset': control_params_offset,
                        'global_params_offset': global_params_offset,
                        'states_offset': states_offset,
                        'controls_offset': controls_offset,
                        'nx': stage.nx,
                        'nu': stage.nu,
                        'ngI': eqI.shape[0],
                        'ng': ng,
                        'ngF': eqF.shape[0],
                        'ng_ineqI': ngIneqI,
                        'ng_ineq': ngIneq,
                        'ng_ineqF': ineqF.shape[0],
                        'n_stage_params': self.stage_params_sym.shape[0],
                        'n_global_params': self.global_params_sym.shape[0],
                        'global_params': global_params_value.T.full().flatten().tolist(), #TODO
                        'stage_params' : stage_params_value.T.full().flatten().tolist(), #TODO
                        'params': [], #TODO
                        'K': self.N+1,
                        'initial_x': X0.T.full().flatten().tolist(),
                        'initial_u': U0.T.full().flatten().tolist(),
                        'lower': cs.repmat(evalf(lb_mid), (self.N-1, 1)).full().flatten().tolist(),
                        'upper': cs.repmat(evalf(ub_mid), (self.N-1, 1)).full().flatten().tolist(),
                        'lowerI': evalf(lb_init).full().flatten().tolist(),
                        'upperI': evalf(ub_init).full().flatten().tolist(),                                               
                        'lowerF': evalf(lb_term).full().flatten().tolist(),
                        'upperF': evalf(ub_term).full().flatten().tolist(),
                        'samplers': list(self.samplers.keys())
                    }
        

        import json
        with open(self.build_dir_abs + os.sep + 'casadi_codegen.json', 'w') as fp:
            json.dump(json_dict, fp, indent = 4)


        class MacroGen:
            def __init__(self):
                self.str = ""
            def process_macro_var(self, macro_name, value):
                self.str += "#define MACRO_" + macro_name + " " + str(value) + "\n"
            def process_macro_array(self, macro_name, value_array):
                self.str += "#define MACRO_" + macro_name + " {" + ", ".join([str(value) if value!=inf else "std::numeric_limits<double>::infinity()" for value in value_array]) + "}\n"
            def __str__(self):
                return self.str
            
        macro_gen = MacroGen()
        macro_gen.process_macro_var('nx', stage.nx)
        macro_gen.process_macro_var('nu', stage.nu)
        macro_gen.process_macro_var('ngI', eqI.shape[0])
        macro_gen.process_macro_var('ng', ng)
        macro_gen.process_macro_var('ngF', eqF.shape[0])
        macro_gen.process_macro_var('ng_ineqI', ngIneqI)
        macro_gen.process_macro_var('ng_ineq', ngIneq)
        macro_gen.process_macro_var('ng_ineqF', ineqF.shape[0])
        macro_gen.process_macro_var('n_stage_params', self.stage_params_sym.shape[0])
        macro_gen.process_macro_var('n_global_params', self.global_params_sym.shape[0])
        macro_gen.process_macro_array('global_params', global_params_value.T.full().flatten().tolist())
        macro_gen.process_macro_array('stage_params' , stage_params_value.T.full().flatten().tolist())
        macro_gen.process_macro_var('K', self.N+1)
        macro_gen.process_macro_array('initial_x', X0.T.full().flatten().tolist())
        macro_gen.process_macro_array('initial_u', U0.T.full().flatten().tolist())
        macro_gen.process_macro_array('bounds_L', evalf(lb_init).full().flatten().tolist() + cs.repmat(evalf(lb_mid), (self.N-1, 1)).full().flatten().tolist() +  evalf(lb_term).full().flatten().tolist())
        macro_gen.process_macro_array('bounds_U',  evalf(ub_init).full().flatten().tolist() + cs.repmat(evalf(ub_mid), (self.N-1, 1)).full().flatten().tolist() +  evalf(ub_term).full().flatten().tolist())
        macro_gen.process_macro_array('samplers', list(self.samplers.keys()))

        # write out the macro_gen to a file
        with open(self.build_dir_abs + '/problem_information.h', 'w') as fp:
            fp.write(str(macro_gen))

        functions_c = str(self.build_dir_abs + "/casadi_codegen.c")
        json_spec = str(self.build_dir_abs + "/casadi_codegen.json")

        if self.mode == "fatropy":
            import fatropy
            if platform.system()=="Windows":
                functions = str(self.build_dir_abs + "/casadi_codegen.dll")
                cmd = f'"{vcvars}" && cl /LD /O2 /Fe: ' + functions + ' ' +  functions_c
                print(cmd)
                subprocess.run(cmd,shell=True)
            else:
                functions = str(self.build_dir_abs + "/casadi_codegen.so")
                subprocess.run("gcc -fPIC -march=native -shared -Ofast " + functions_c + " -o " + functions, shell = True)
            # self.myOCP = fatropy.OCP(functions,json_spec)
            self.myOCP = fatropy.StageOCPApplicationFactory.from_rockit_interface(functions, json_spec)

        elif self.mode == "interface":
            # check if interface_generation directory exists
            if not os.path.exists(self.build_dir_abs + "/interface/build"):
                # copy the interface_generation directory of this files path to the current directory
                shutil.copytree(os.path.dirname(os.path.realpath(__file__)) + os.sep + "interface_generation",self.build_dir_abs + "/interface/")
            
            with open(self.build_dir_abs + "/interface/after_init.c.in", "w") as after_init:
                for name, value in self.fatrop_options.items():
                    after_init.write(f"""set_option("{name}",{value});\n""")
            # run cmake to build the interface
            subprocess.run(["cmake","-DCMAKE_BUILD_TYPE=Release","-DWITH_PYTHON=OFF","-DBUILD_EXECUTABLES=OFF","-S", "interface","-B", os.path.join("interface","build")], cwd=self.build_dir_abs)
            # run make to build the interface
            subprocess.run(["cmake","--build","interface/build","--config","Release"], cwd=self.build_dir_abs)
            subprocess.run(["cmake","--install","interface/build","--prefix","."], cwd=self.build_dir_abs)
            self.fatrop_solver = cs.external("fatrop_driver", self.build_dir_abs + "/lib/libfatrop_driver.so")
            # self.solver.checkout()
            # use ctypes to call
            # void set_option(const std::string& option, const double& value)
            # from libfatrop_driver.so
            import ctypes
            # Declare the function prototype
            self.lib = ctypes.PyDLL(self.build_dir_abs + "/lib/libfatrop_driver.so")
            set_option_argtypes = [ctypes.c_char_p, ctypes.c_double]
            self._register("set_option", set_option_argtypes, None)
            pass 

        return
    
    def set_option(self, option, value):
        if self.mode == "fatropy":
            self.myOCP.set_option(option, value)
        elif self.mode == "interface":
            self._set_option(option.encode('utf-8'), value)
        return
    
    def solver(self, name , opts: dict):
        for key, value in opts.items():
            self.set_option(key, value)
        return




    def to_function(self, stage, name, args, results, *margs):
        args = [stage.value(a) for a in args]
        class FatropCallBack(Callback):
            def __init__(self,solver, n_stage_params, n_global_params,nu, nx, K):
                Callback.__init__(self)
                self.fatrop_solver = solver
                self.nx = nx 
                self.nu = nu 
                self.K = K 
                self.n_stage_params = n_stage_params
                self.n_global_params = n_global_params
                self.construct("fatrop_callback", {})
            def get_n_in(self):
                return 4 # stage_params, global_params, u, x
            def get_n_out(self):
                return 2 # u_opt, x_opt
            def get_sparsity_in(self, input):
                if(input ==0):
                    return cs.Sparsity.dense(self.n_stage_params, self.K) 
                if(input ==1):
                    return cs.Sparsity.dense(self.n_global_params, 1) 
                if(input ==2):
                    return cs.Sparsity.dense(self.nu, self.K-1) 
                if(input ==3):
                    return cs.Sparsity.dense(self.nx, self.K) 
                return
                # return super().get_sparsity_in(*args)
            def get_sparsity_out(self, output):
                if(output ==0):
                    return cs.Sparsity.dense(self.nu, self.K-1) 
                if(output ==1):
                    return cs.Sparsity.dense(self.nx, self.K) 
                return
            def init(self):
                return
            
            def eval(self, arg):
                self.fatrop_solver.set_params(arg[0], arg[1])
                self.fatrop_solver.set_initial_u(cs.DM(arg[2]))
                self.fatrop_solver.set_initial_x(cs.DM(arg[3]))
                self.fatrop_solver.optimize()
                sol = self.fatrop_solver.last_solution()
                return [cs.DM(sol.u), cs.DM(sol.x)] 
            
            def has_codegen(sefl):
                return True
            
        controls = cs.hcat(self.U_gist)
        states = cs.hcat(self.X_gist)
        p_stage = cs.hcat(self.P_stage_gist)
        p_global = self.P_global_gist

        gist_list = self.U_gist+self.X_gist+self.P_stage_gist+[self.P_global_gist,self.T_gist]

        try:
            helper0 = Function("helper0", args, gist_list, {"allow_free":True})
        except:
            helper0 = Function("helper0", args, gist_list)

        if helper0.has_free():
            helper0_free = helper0.free_mx()
            print("test",helper0_free, self.initial_value(stage, helper0_free))
            [controls, states, p_stage, p_global, T_gist] = cs.substitute([controls, states, p_stage, p_global, self.T_gist], helper0_free, self.initial_value(stage, helper0_free))
        if self.mode == "fatropy":
            self.fatrop_solver = FatropCallBack(self.myOCP, self.P_stage_gist[0].size1(), self.P_global_gist.size1(), stage.nu, stage.nx, self.N + 1)
            res_u, res_x = self.fatrop_solver(p_stage, p_global, controls, states)
        elif self.mode == "interface":
            res_x, res_u, _, _ = self.fatrop_solver(densify(p_global.monitor("p_global")), densify(p_stage.monitor("p_stage")), densify(states), densify(controls))

        ret = Function(name, args, cs.substitute(results, gist_list, cs.horzsplit(res_u)+cs.horzsplit(res_x)+cs.horzsplit(p_stage)+[p_global,T_gist]) , *margs)
        assert not ret.has_free()
        return ret

    def initial_value(self, stage, expr):
        # check if expr is a list
        lst = isinstance(expr, list) 
        if not lst:
            expr = [expr]
        # expr = stage.value(expr)
        [_,states] = stage.sample(stage.x,grid='control')
        [_,controls] = stage.sample(stage.u,grid='control-')
        variables = stage.value(veccat(stage.variables['']))

        helper_in = [self.P_global_gist,cs.horzcat(*self.P_stage_gist),states,controls,variables,self.T_gist]
        helper = Function("helper", helper_in, expr)
        ret = [s.toarray(simplify=True) for s in helper.call([self.global_params_value, self.stage_params_value, self.X0, self.U0, 0, self.inits[0][1]])]
        if not lst :
            return ret[0]
        return ret

    def solve(self, stage,limited=False):
        # self.global_params_value, self.stage_params_value = self.get_parameters(stage)
        if self.mode == "fatropy":
            self.myOCP.set_params(self.stage_params_value, self.global_params_value)
            retval = self.myOCP.optimize()
            sol = self.myOCP.last_solution()
            u_sol_fatrop = sol.u 
            x_sol_fatrop = sol.x
        elif self.mode == "interface":
            x_sol_fatrop, u_sol_fatrop, _, _= self.fatrop_solver(self.global_params_value, self.stage_params_value, self.X0, self.U0)



        return OcpSolution(SolWrapper(self, vec(x_sol_fatrop), veccat(vec(u_sol_fatrop), DM.zeros(1)), DM.zeros(0), DM.zeros(0),cs.vertcat(self.stage_params_value), cs.vertcat(self.global_params_value), rT=stage.T), stage)

    def get_stats(self):
        stats = self.stats
        return dict((k,getattr(stats,k)) for k,_ in stats_fields)

    def non_converged_solution(self, stage):
        return self.last_solution

    def solve_limited(self, stage):
        return self.solve(stage,limited=True)

    def eval(self, stage, expr):
        placeholders = stage.placeholders_transcribed
        expr = placeholders(expr,max_phase=1)
        ks = [cs.vvcat(stage.parameters['']), stage.T]
        vs = [self.P_global_gist, self.T_gist]
        ret = substitute([expr],ks,vs)[0]
        return ret
        
    def eval_at_control(self, stage, expr, k):
        placeholders = stage.placeholders_transcribed
        expr = placeholders(expr,max_phase=1)
        ks = [stage.x,stage.u,self.p_global,stage.T, cs.vvcat(stage.parameters['control'] + stage.parameters['control+'])]
        vs = [self.X_gist[k], self.U_gist[min(k, self.N-1)], self.P_global_gist, self.T_gist, self.P_stage_gist[k]]
        if not self.t_state:
            ks += [stage.t]
            vs += [self.control_grid[k]]
        ret = substitute([expr],ks,vs)[0]
        return ret

class SolWrapper:
    def __init__(self, method, x, u, v, T,P_stage, P_global, rT=None):
        self.method = method
        self.x = x
        self.u = u
        self.T = T
        self.v = v
        self.rT = rT
        self.P_stage = P_stage
        self.P_global = P_global

    def value(self, expr, *args,**kwargs):
        placeholders = self.method.stage.placeholders_transcribed
        expr = substitute(expr,self.rT, self.T)
        ret = evalf(substitute([placeholders(expr)],[self.method.gist],[veccat(self.x, self.u, self.v,self.P_stage, self.P_global ,self.T)])[0])
        return ret.toarray(simplify=True)

    def stats(self):
        return self.method.get_stats()



# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

import importlib
import json
import os
import sys
from copy import deepcopy
from ctypes import (POINTER, byref, c_bool, c_char_p, c_double, c_int,
                    c_void_p, cast)
if os.name == 'nt':
    from ctypes import wintypes
    from ctypes import WinDLL as DllLoader
else:
    from ctypes import CDLL as DllLoader

import numpy as np

from .acados_model import AcadosModel
from .acados_ocp import AcadosOcp
from .acados_sim import AcadosSim, AcadosSimDims, AcadosSimOptions

from .builders import CMakeBuilder
from .casadi_function_generation import (
                    GenerateContext,
                    generate_c_code_explicit_ode,
                    generate_c_code_gnsf,
                    generate_c_code_implicit_ode)
from .gnsf.detect_gnsf_structure import detect_gnsf_structure
from .utils import (check_casadi_version, format_class_dict,
                    get_shared_lib_ext, get_shared_lib_prefix, get_shared_lib_dir,
                    make_object_json_dumpable,
                    render_template, set_up_imported_gnsf_model,
                    verbose_system_call, acados_lib_is_compiled_with_openmp,
                    get_shared_lib)


def sim_formulation_json_dump(acados_sim: AcadosSim, json_file='acados_sim.json'):

    # Copy input sim object dictionary
    sim_dict = dict(deepcopy(acados_sim).__dict__)

    # convert acados classes to dicts
    for key, v in sim_dict.items():
        # skip non dict attributes
        if isinstance(v, (AcadosSim, AcadosSimDims, AcadosSimOptions, AcadosModel)):
            sim_dict[key]=dict(getattr(acados_sim, key).__dict__)

    sim_json = format_class_dict(sim_dict)

    with open(json_file, 'w') as f:
        json.dump(sim_json, f, default=make_object_json_dumpable, indent=4, sort_keys=True)



def sim_render_templates(json_file, model_name: str, code_export_dir, cmake_options: CMakeBuilder = None):
    # setting up loader and environment
    json_path = os.path.join(os.getcwd(), json_file)

    if not os.path.exists(json_path):
        raise Exception(f"{json_path} not found!")

    # Render templates
    in_file = 'acados_sim_solver.in.c'
    out_file = f'acados_sim_solver_{model_name}.c'
    render_template(in_file, out_file, code_export_dir, json_path)

    in_file = 'acados_sim_solver.in.h'
    out_file = f'acados_sim_solver_{model_name}.h'
    render_template(in_file, out_file, code_export_dir, json_path)

    in_file = 'acados_sim_solver.in.pxd'
    out_file = f'acados_sim_solver.pxd'
    render_template(in_file, out_file, code_export_dir, json_path)

    # Builder
    if cmake_options is not None:
        in_file = 'CMakeLists.in.txt'
        out_file = 'CMakeLists.txt'
        render_template(in_file, out_file, code_export_dir, json_path)
    else:
        in_file = 'Makefile.in'
        out_file = 'Makefile'
        render_template(in_file, out_file, code_export_dir, json_path)

    in_file = 'main_sim.in.c'
    out_file = f'main_sim_{model_name}.c'
    render_template(in_file, out_file, code_export_dir, json_path)

    # folder model
    model_dir = os.path.join(code_export_dir, model_name + '_model')

    in_file = 'model.in.h'
    out_file = f'{model_name}_model.h'
    render_template(in_file, out_file, model_dir, json_path)


def sim_generate_external_functions(acados_sim: AcadosSim):
    model = acados_sim.model

    integrator_type = acados_sim.solver_options.integrator_type

    opts = dict(generate_hess = acados_sim.solver_options.sens_hess,
                code_export_directory = acados_sim.code_export_directory)

    # create code_export_dir, model_dir
    code_export_dir = acados_sim.code_export_directory
    opts['code_export_directory'] = code_export_dir
    model_dir = os.path.join(code_export_dir, model.name + '_model')
    if not os.path.exists(model_dir):
        os.makedirs(model_dir)

    context = GenerateContext(model.p_global, model.name, opts)

    # generate external functions
    check_casadi_version()
    if integrator_type == 'ERK':
        generate_c_code_explicit_ode(context, model, model_dir)
    elif integrator_type == 'IRK':
        generate_c_code_implicit_ode(context, model, model_dir)
    elif integrator_type == 'GNSF':
        generate_c_code_gnsf(context, model, model_dir)

    context.finalize()
    acados_sim.__external_function_files_model = context.get_external_function_file_list(ocp_specific=False)


class AcadosSimSolver:
    """
    Class to interact with the acados integrator C object.

        :param acados_sim: type :py:class:`~acados_template.acados_ocp.AcadosOcp` (takes values to generate an instance :py:class:`~acados_template.acados_sim.AcadosSim`) or :py:class:`~acados_template.acados_sim.AcadosSim`
        :param json_file: Default: 'acados_sim.json'
        :param build: Default: True
        :param cmake_builder: type :py:class:`~acados_template.utils.CMakeBuilder` generate a `CMakeLists.txt` and use
            the `CMake` pipeline instead of a `Makefile` (`CMake` seems to be the better option in conjunction with
            `MS Visual Studio`); default: `None`
    """
    if sys.platform=="win32":
        dlclose = DllLoader('kernel32', use_last_error=True).FreeLibrary
        dlclose.argtypes = [wintypes.HMODULE]
        winmode = 8 # why 8? what does that mean?
    else:
        dlclose = DllLoader(None).dlclose
        dlclose.argtypes = [c_void_p]
        winmode = None

    @property
    def acados_lib_uses_omp(self,):
        """`acados_lib_uses_omp` - flag indicating whether the acados library has been compiled with openMP."""
        return self.__acados_lib_uses_omp

    @classmethod
    def generate(cls, acados_sim: AcadosSim, json_file='acados_sim.json', cmake_builder: CMakeBuilder = None):
        """
        Generates the code for an acados sim solver, given the description in acados_sim
        """

        acados_sim.code_export_directory = os.path.abspath(acados_sim.code_export_directory)

        acados_sim.make_consistent()

        # module dependent post processing
        if acados_sim.solver_options.integrator_type == 'GNSF':
            if acados_sim.solver_options.sens_hess == True:
                raise Exception("AcadosSimSolver: GNSF does not support sens_hess = True.")
            if 'gnsf_model' in acados_sim.__dict__:
                set_up_imported_gnsf_model(acados_sim)
            else:
                detect_gnsf_structure(acados_sim)

        # generate external functions
        sim_generate_external_functions(acados_sim)

        # dump to json
        sim_formulation_json_dump(acados_sim, json_file)

        # render templates
        sim_render_templates(json_file, acados_sim.model.name, acados_sim.code_export_directory, cmake_builder)


    @classmethod
    def build(cls, code_export_dir, with_cython=False, cmake_builder: CMakeBuilder = None, verbose: bool = True):
        # Compile solver
        cwd = os.getcwd()
        os.chdir(code_export_dir)
        if with_cython:
            verbose_system_call(['make', 'clean_sim_cython'], verbose)
            verbose_system_call(['make', 'sim_cython'], verbose)
        else:
            if cmake_builder is not None:
                cmake_builder.exec(code_export_dir, verbose)
            else:
                verbose_system_call(['make', 'sim_shared_lib'], verbose)
        os.chdir(cwd)


    @classmethod
    def create_cython_solver(cls, json_file):
        """
        """
        with open(json_file, 'r') as f:
            acados_sim_json = json.load(f)
        code_export_directory = acados_sim_json['code_export_directory']

        importlib.invalidate_caches()
        sys.path.append(os.path.dirname(code_export_directory))
        acados_sim_solver_pyx = importlib.import_module(f'{os.path.split(code_export_directory)[1]}.acados_sim_solver_pyx')

        AcadosSimSolverCython = getattr(acados_sim_solver_pyx, 'AcadosSimSolverCython')
        return AcadosSimSolverCython(acados_sim_json['model']['name'])

    def __init__(self, acados_sim: AcadosSim, json_file='acados_sim.json', generate=True, build=True, cmake_builder: CMakeBuilder = None, verbose: bool = True):

        self.solver_created = False
        model_name = acados_sim.model.name
        self.model_name = model_name

        code_export_dir = os.path.abspath(acados_sim.code_export_directory)

        # reuse existing json and casadi functions, when creating integrator from ocp
        if generate and not isinstance(acados_sim, AcadosOcp):
            self.generate(acados_sim, json_file=json_file, cmake_builder=cmake_builder)

        if isinstance(acados_sim, AcadosOcp):
            print("Warning: An AcadosSimSolver is created from an AcadosOcp description.",
                  "This only works if you created an AcadosOcpSolver before with the same description."
                  "Otherwise it leads to undefined behavior. Using an AcadosSim description is recommended.")
            self.T = acados_sim.solver_options.Tsim
        else:
            self.T = acados_sim.solver_options.T

        self.acados_sim = acados_sim

        if build:
            self.build(code_export_dir, cmake_builder=cmake_builder, verbose=verbose)

        # prepare library loading
        lib_ext = get_shared_lib_ext()
        lib_prefix = get_shared_lib_prefix()
        lib_dir = get_shared_lib_dir()

        # Load acados library to avoid unloading the library.
        # This is necessary if acados was compiled with OpenMP, since the OpenMP threads can't be destroyed.
        # Unloading a library which uses OpenMP results in a segfault (on any platform?).
        # see [https://stackoverflow.com/questions/34439956/vc-crash-when-freeing-a-dll-built-with-openmp]
        # or [https://python.hotexamples.com/examples/_ctypes/-/dlclose/python-dlclose-function-examples.html]
        libacados_name = f'{lib_prefix}acados{lib_ext}'
        libacados_filepath = os.path.join(acados_sim.acados_lib_path, '..', lib_dir, libacados_name)
        self.__acados_lib = get_shared_lib(libacados_filepath, self.winmode)

        # find out if acados was compiled with OpenMP
        self.__acados_lib_uses_omp = acados_lib_is_compiled_with_openmp(self.__acados_lib, verbose)

        libacados_sim_solver_name = f'{lib_prefix}acados_sim_solver_{self.model_name}{lib_ext}'
        self.shared_lib_name = os.path.join(code_export_dir, libacados_sim_solver_name)
        self.shared_lib = get_shared_lib(self.shared_lib_name, winmode=self.winmode)

        # create capsule
        getattr(self.shared_lib, f"{model_name}_acados_sim_solver_create_capsule").restype = c_void_p
        self.capsule = getattr(self.shared_lib, f"{model_name}_acados_sim_solver_create_capsule")()

        # create solver
        getattr(self.shared_lib, f"{model_name}_acados_sim_create").argtypes = [c_void_p]
        getattr(self.shared_lib, f"{model_name}_acados_sim_create").restype = c_int
        assert getattr(self.shared_lib, f"{model_name}_acados_sim_create")(self.capsule)==0
        self.solver_created = True

        getattr(self.shared_lib, f"{model_name}_acados_get_sim_opts").argtypes = [c_void_p]
        getattr(self.shared_lib, f"{model_name}_acados_get_sim_opts").restype = c_void_p
        self.sim_opts = getattr(self.shared_lib, f"{model_name}_acados_get_sim_opts")(self.capsule)

        getattr(self.shared_lib, f"{model_name}_acados_get_sim_dims").argtypes = [c_void_p]
        getattr(self.shared_lib, f"{model_name}_acados_get_sim_dims").restype = c_void_p
        self.sim_dims = getattr(self.shared_lib, f"{model_name}_acados_get_sim_dims")(self.capsule)

        getattr(self.shared_lib, f"{model_name}_acados_get_sim_config").argtypes = [c_void_p]
        getattr(self.shared_lib, f"{model_name}_acados_get_sim_config").restype = c_void_p
        self.sim_config = getattr(self.shared_lib, f"{model_name}_acados_get_sim_config")(self.capsule)

        getattr(self.shared_lib, f"{model_name}_acados_get_sim_out").argtypes = [c_void_p]
        getattr(self.shared_lib, f"{model_name}_acados_get_sim_out").restype = c_void_p
        self.sim_out = getattr(self.shared_lib, f"{model_name}_acados_get_sim_out")(self.capsule)

        getattr(self.shared_lib, f"{model_name}_acados_get_sim_in").argtypes = [c_void_p]
        getattr(self.shared_lib, f"{model_name}_acados_get_sim_in").restype = c_void_p
        self.sim_in = getattr(self.shared_lib, f"{model_name}_acados_get_sim_in")(self.capsule)

        getattr(self.shared_lib, f"{model_name}_acados_get_sim_solver").argtypes = [c_void_p]
        getattr(self.shared_lib, f"{model_name}_acados_get_sim_solver").restype = c_void_p
        self.sim_solver = getattr(self.shared_lib, f"{model_name}_acados_get_sim_solver")(self.capsule)

        self.gettable_vectors = ['x', 'u', 'z', 'S_adj']
        self.gettable_matrices = ['S_forw', 'Sx', 'Su', 'S_hess', 'S_algebraic']
        self.gettable_scalars = ['CPUtime', 'time_tot', 'ADtime', 'time_ad', 'LAtime', 'time_la']


    def simulate(self, x=None, u=None, z=None, xdot=None, p=None):
        """
        Simulate the system forward for the given x, u, p and return x_next.
        The values xdot, z are used as initial guesses for implicit integrators, if provided.
        Wrapper around `solve()` taking care of setting/getting inputs/outputs.
        """
        if x is not None:
            self.set('x', x)
        if u is not None:
            self.set('u', u)
        if self.acados_sim.solver_options.integrator_type == "IRK":
            if z is not None:
                self.set('z', z)
            if xdot is not None:
                self.set('xdot', xdot)
        if p is not None:
            self.set('p', p)

        status = self.solve()

        if status != 0:
            raise Exception(f'acados_sim_solver for model {self.model_name} returned status {status}.')

        x_next = self.get('x')
        return x_next


    def solve(self):
        """
        Solve the simulation problem with current input.
        """
        getattr(self.shared_lib, f"{self.model_name}_acados_sim_solve").argtypes = [c_void_p]
        getattr(self.shared_lib, f"{self.model_name}_acados_sim_solve").restype = c_int

        status = getattr(self.shared_lib, f"{self.model_name}_acados_sim_solve")(self.capsule)
        return status


    def get(self, field_):
        """
        Get the last solution of the solver.

            :param str field: string in ['x', 'u', 'z', 'S_forw', 'Sx', 'Su', 'S_adj', 'S_hess', 'S_algebraic', 'CPUtime', 'time_tot', 'ADtime', 'time_ad', 'LAtime', 'time_la']
        """
        field = field_.encode('utf-8')

        if field_ in self.gettable_vectors:
            # get dims
            dims = np.ascontiguousarray(np.zeros((2,)), dtype=np.intc)
            dims_data = cast(dims.ctypes.data, POINTER(c_int))

            self.__acados_lib.sim_dims_get_from_attr.argtypes = [c_void_p, c_void_p, c_char_p, POINTER(c_int)]
            self.__acados_lib.sim_dims_get_from_attr(self.sim_config, self.sim_dims, field, dims_data)

            # allocate array
            out = np.ascontiguousarray(np.zeros((dims[0],)), dtype=np.float64)
            out_data = cast(out.ctypes.data, POINTER(c_double))

            self.__acados_lib.sim_out_get.argtypes = [c_void_p, c_void_p, c_void_p, c_char_p, c_void_p]
            self.__acados_lib.sim_out_get(self.sim_config, self.sim_dims, self.sim_out, field, out_data)

        elif field_ in self.gettable_matrices:
            # get dims
            dims = np.ascontiguousarray(np.zeros((2,)), dtype=np.intc)
            dims_data = cast(dims.ctypes.data, POINTER(c_int))

            self.__acados_lib.sim_dims_get_from_attr.argtypes = [c_void_p, c_void_p, c_char_p, POINTER(c_int)]
            self.__acados_lib.sim_dims_get_from_attr(self.sim_config, self.sim_dims, field, dims_data)

            out = np.zeros((dims[0], dims[1]), order='F')
            out_data = cast(out.ctypes.data, POINTER(c_double))

            self.__acados_lib.sim_out_get.argtypes = [c_void_p, c_void_p, c_void_p, c_char_p, c_void_p]
            self.__acados_lib.sim_out_get(self.sim_config, self.sim_dims, self.sim_out, field, out_data)

        elif field_ in self.gettable_scalars:
            scalar = c_double()
            scalar_data = byref(scalar)
            self.__acados_lib.sim_out_get.argtypes = [c_void_p, c_void_p, c_void_p, c_char_p, c_void_p]
            self.__acados_lib.sim_out_get(self.sim_config, self.sim_dims, self.sim_out, field, scalar_data)

            out = scalar.value
        else:
            raise Exception(f'AcadosSimSolver.get(): Unknown field {field_},' \
                f' available fields are {", ".join(self.gettable_vectors+self.gettable_matrices)}, {", ".join(self.gettable_scalars)}')

        return out



    def set(self, field_: str, value_):
        """
        Set numerical data inside the solver.

            :param field: string in ['x', 'u', 'p', 'xdot', 'z', 'seed_adj', 'T', 't0']
            :param value: the value with appropriate size.
        """
        settable = ['x', 'u', 'p', 'xdot', 'z', 'seed_adj', 'T', 't0'] # S_forw

        # TODO: check and throw error here. then remove checks in Cython for speed
        # cast value_ to avoid conversion issues
        if isinstance(value_, (float, int)):
            value_ = np.array([value_])

        value_ = value_.astype(float)
        value_data = cast(value_.ctypes.data, POINTER(c_double))
        value_data_p = cast((value_data), c_void_p)

        field = field_.encode('utf-8')

        # treat parameters separately
        if field_ == 'p':
            model_name = self.acados_sim.model.name
            getattr(self.shared_lib, f"{model_name}_acados_sim_update_params").argtypes = [c_void_p, POINTER(c_double), c_int]
            value_data = cast(value_.ctypes.data, POINTER(c_double))
            getattr(self.shared_lib, f"{model_name}_acados_sim_update_params")(self.capsule, value_data, value_.shape[0])
            return
        else:
            # dimension check
            dims = np.ascontiguousarray(np.zeros((2,)), dtype=np.intc)
            dims_data = cast(dims.ctypes.data, POINTER(c_int))

            self.__acados_lib.sim_dims_get_from_attr.argtypes = [c_void_p, c_void_p, c_char_p, POINTER(c_int)]
            self.__acados_lib.sim_dims_get_from_attr(self.sim_config, self.sim_dims, field, dims_data)

            value_ = np.ravel(value_, order='F')

            value_shape = value_.shape
            if len(value_shape) == 1:
                value_shape = (value_shape[0], 0)

            if value_shape != tuple(dims):
                raise Exception(f'AcadosSimSolver.set(): mismatching dimension' \
                    f' for field "{field_}" with dimension {tuple(dims)} (you have {value_shape}).')

            if field_ == 'T':
                self.T = value_

        # set
        if field_ in ['xdot', 'z']:
            self.__acados_lib.sim_solver_set.argtypes = [c_void_p, c_char_p, c_void_p]
            self.__acados_lib.sim_solver_set(self.sim_solver, field, value_data_p)
        elif field_ in settable:
            self.__acados_lib.sim_in_set.argtypes = [c_void_p, c_void_p, c_void_p, c_char_p, c_void_p]
            self.__acados_lib.sim_in_set(self.sim_config, self.sim_dims, self.sim_in, field, value_data_p)
        else:
            raise Exception(f'AcadosSimSolver.set(): Unknown field {field_},' \
                f' available fields are {", ".join(settable)}')

        return


    def options_set(self, field_: str, value_: bool):
        """
        Set solver options

            :param field: string in ['sens_forw', 'sens_adj', 'sens_hess']
            :param value: Boolean
        """
        fields = ['sens_forw', 'sens_adj', 'sens_hess']
        if field_ not in fields:
            raise Exception(f"field {field_} not supported. Supported values are {', '.join(fields)}.\n")

        field = field_.encode('utf-8')
        value_ctypes = c_bool(value_)

        if not isinstance(value_, bool):
            raise TypeError("options_set: expected boolean for value")

        # only allow setting
        if getattr(self.acados_sim.solver_options, field_) or value_ == False:
            self.__acados_lib.sim_opts_set.argtypes = [c_void_p, c_void_p, c_char_p, POINTER(c_bool)]
            self.__acados_lib.sim_opts_set(self.sim_config, self.sim_opts, field, value_ctypes)
        else:
            raise RuntimeError(f"Cannot set option {field_} to True, because it was False in original solver options.\n")

        return


    def __del__(self):

        if self.solver_created:
            getattr(self.shared_lib, f"{self.model_name}_acados_sim_free").argtypes = [c_void_p]
            getattr(self.shared_lib, f"{self.model_name}_acados_sim_free").restype = c_int
            getattr(self.shared_lib, f"{self.model_name}_acados_sim_free")(self.capsule)

            getattr(self.shared_lib, f"{self.model_name}_acados_sim_solver_free_capsule").argtypes = [c_void_p]
            getattr(self.shared_lib, f"{self.model_name}_acados_sim_solver_free_capsule").restype = c_int
            getattr(self.shared_lib, f"{self.model_name}_acados_sim_solver_free_capsule")(self.capsule)

            try:
                self.dlclose(self.shared_lib._handle)
            except:
                print(f"WARNING: acados Python interface could not close shared_lib handle of AcadosSimSolver {self.model_name}.\n",
                     "Attempting to create a new one with the same name will likely result in the old one being used!")
                pass
